"""
Tools to help interface the likelihood code with other software.

Overview:
---------------

Theory Code Interface Functions:

* :func:`get_CosmoPowerJAX_pars_to_theory_specs_func`
* :func:`get_CosmoPower_pars_to_theory_specs_func`
* :func:`get_PyCapse_pars_to_theory_specs_func`
* :func:`get_CAMB_pars_to_theory_specs_func`
* :func:`get_CLASS_pars_to_theory_specs_func`
* :func:`get_CobayaTheory_pars_to_theory_specs_func`

MontePython Interface:

* :func:`get_montepython_nuisance_param_block_for_like`

Cobaya Interface:

* :class:`CandlCobayaLikelihood`
* :func:`get_cobaya_likelihood_class_for_like`
* :func:`get_cobaya_info_dict_for_like`

Cobaya Theory Code Classes:

* :class:`CobayaTheoryCosmoPowerJAX`
* :class:`CobayaTheoryCosmoPower`
* :class:`CobayaTheoryPyCapse`
* :class:`CobayaTheoryBBTemplate`
* :class:`CobayaTheoryCosmoPowerJAXLensing`
* :class:`CobayaTheoryCosmoPowerLensing`
"""

# --------------------------------------#
# IMPORTS
# --------------------------------------#

from candl.lib import *
import candl.io
import candl

# --------------------------------------#
# COSMOPOWER - COBAYA INTERFACE
# --------------------------------------#


class CobayaTheoryCosmoPowerJAX(cobaya_theory_Theory):
    """
    Wraps CosmoPower-JAX model into a cobaya.theory.Theory class.
    See D. Piras, A. Spurio Mancini 2023 and A. Spurio Mancini et al. 2021 for more (https://arxiv.org/abs/2305.06347, https://arxiv.org/abs/2106.03846).

    To initialise, pass 'emulator_filenames' in the relevant cobaya block:
    A dictionary of spectrum types and emulator file names (if they are placed in cosmopower-jax's trained_models/ folder), or the full path of the files, including in either case the file ending '.pkl' or '.npz'.

    This code is taken from the Cobaya example for custom theory codes and only slightly modified.
    See https://cobaya.readthedocs.io/en/latest/theories_and_dependencies.html.
    Torrado and Lewis, 2020 (https://arxiv.org/abs/2005.05290)

    Attributes
    ----------------
    emulator_filenames : dict
        File names of emulators.
    cp_emulators : dict
        CosmoPower emulators.
    cp_pars : list
        List of parameters required by the emulators.
    provider : Provider
        Cobaya provider.
    current_state : dict
        Dict containing current parameters and results.

    Methods
    ---------
    initialize :
        Initialises an instance of the class (loads emulators).
    initialize_with_provider :
        Initialization after other components initialised.
    get_requirements :
        Returns what the theory code needs to run.
    must_provide :
        Returns what the theory code needs to run.
    get_can_provide :
        Return what the theory code can supply.
    calculate :
        Carry out the calculation.
    get_Cl :
        Return result of calculation (Cl and Dl).
    """

    # I know having this be a class variable is not super elegant, but it's the easiest way to interface with Cobaya
    emulator_filenames: dict = {}

    def initialize(self):
        """
        Called from __init__ to initialise. Loads the emulator models.

        Notes
        ----------
        For CosmoPower-JAX emulator models are expected to be placed in the package directory.
        TE emulators are loaded as PCA+NN, other spectra as NN-only models.

        """

        self.cp_emulators = {}
        for spec_type in list(self.emulator_filenames.keys()):
            if spec_type == "TE":
                try:
                    self.cp_emulators[spec_type] = CPJ(
                        probe="custom_pca",
                        filename=self.emulator_filenames[spec_type],
                    )
                except:
                    self.cp_emulators[spec_type] = CPJ(
                        probe="custom_pca",
                        filepath=self.emulator_filenames[spec_type],
                    )
            else:
                try:
                    self.cp_emulators[spec_type] = CPJ(
                        probe="custom_log",
                        filename=self.emulator_filenames[spec_type],
                    )
                except:
                    self.cp_emulators[spec_type] = CPJ(
                        probe="custom_log",
                        filepath=self.emulator_filenames[spec_type],
                    )

        # Just using a random spec here, would be weird if TT/TE/EE had different requirements
        self.cp_pars = list(list(self.cp_emulators.values())[0].parameters)
        if "h" in self.cp_pars:
            self.cp_pars[self.cp_pars.index("h")] = "H0"

    def initialize_with_provider(self, provider):
        """
        Initialization after other components initialized, using Provider class
        instance which is used to return any dependencies (see calculate below).
        """

        self.provider = provider

    def get_requirements(self):
        """
        Return dictionary of derived parameters or other quantities that are needed
        by this component and should be calculated by another theory class.
        """
        return {p: None for p in self.cp_pars}

    def must_provide(self, **requirements):
        """
        Return dictionary of parameters that must be provided.
        """
        return {p: None for p in self.cp_pars}

    def get_can_provide(self):
        """
        Return list of quantities that can be provided.
        """
        return ["Cl", "Dl"]

    def calculate(self, state, want_derived=True, **params_values_dict):
        """
        Calculate the CMB spectra.
        Calls the CosmoPower-JAX emulator models.
        """
        # Prepare hand-off to CosmoPower
        # numpy array creation is faster, but need jnp to get derivs
        # This is the order expected by CosmoPower-JAX

        pars_for_cp = [state["params"][p] for p in self.cp_pars]
        if "H0" in self.cp_pars:
            pars_for_cp[self.cp_pars.index("H0")] /= 100
        pars_for_cp = jnp.array(pars_for_cp)

        # Get CMB Cls
        state["Cl"] = {
            "ell": self.cp_emulators[list(self.cp_emulators.keys())[0]].modes
        }  # assume all supply the same ell range}
        state["Dl"] = {"ell": state["Cl"]["ell"]}
        for spec_type in list(self.cp_emulators.keys()):
            state["Cl"][spec_type] = (
                self.cp_emulators[spec_type].predict(pars_for_cp).ravel()
            )
            state["Dl"][spec_type] = (
                state["Cl"][spec_type]
                * self.cp_emulators[spec_type].modes
                * (self.cp_emulators[spec_type].modes + 1)
                / (2 * jnp.pi)
            )

    def get_Cl(self, ell_factor=False, **kwargs):
        """
        Return the Cls or Dls
        """
        if ell_factor:
            return self.current_state["Dl"].copy()
        else:
            return self.current_state["Cl"].copy()


class CobayaTheoryCosmoPower(cobaya_theory_Theory):
    """
    Wraps regular CosmoPower model into a cobaya theory class.
    See A. Spurio Mancini et al. 2021 for more (https://arxiv.org/abs/2106.03846).

    To initialise, pass 'emulator_filenames' in the relevant cobaya block:
    A dictionary of spectrum type (TT, TE, ...) and file names of corresponding emulator models.

    This code is taken from the Cobaya example for custom theory codes and only slightly modified.
    See https://cobaya.readthedocs.io/en/latest/theories_and_dependencies.html.
    Torrado and Lewis, 2020 (https://arxiv.org/abs/2005.05290)

    Attributes
    -------------
    emulator_filenames : dict
        File names of emulators.
    cp_emulators : dict
        CosmoPower emulators.
    cp_pars : list
        List of parameters required by the emulators.
    provider : Provider
        Cobaya provider.
    current_state : dict
        Dict containing current parameters and results.

    Methods
    ---------
    initialize :
        Complete set-up.
    initialize_with_provider :
        Initialization after other components initialised.
    get_requirements :
        Returns what the theory code needs to run.
    must_provide :
        Returns what the theory code needs to run.
    get_can_provide :
        Return what the theory code can supply.
    calculate :
        Carry out the calculation.
    get_Cl :
        Return result of calculation (Cl and Dl).
    """

    # I know having this be a class variable is not super elegant, but it's the easiest way to interface with Cobaya
    emulator_filenames: dict = {}

    def initialize(self):
        """
        Called from __init__ to initialise.
        Loads the emulator models.
        """

        # Load models and unify prediction methods
        self.cp_emulators = {}
        for spec_type in list(self.emulator_filenames.keys()):
            if spec_type == "TE":
                self.cp_emulators[spec_type] = cp.cosmopower_PCAplusNN(
                    restore=True, restore_filename=self.emulator_filenames[spec_type]
                )
                self.cp_emulators[spec_type].get_prediction = self.cp_emulators[
                    spec_type
                ].predictions_np
            else:
                self.cp_emulators[spec_type] = cp.cosmopower_NN(
                    restore=True, restore_filename=self.emulator_filenames[spec_type]
                )
                self.cp_emulators[spec_type].get_prediction = self.cp_emulators[
                    spec_type
                ].ten_to_predictions_np

        # Just using a random spec here, would be weird if TT/TE/EE had different requirements
        self.cp_pars = list(list(self.cp_emulators.values())[0].parameters)
        if "h" in self.cp_pars:
            self.cp_pars[self.cp_pars.index("h")] = "H0"

    def initialize_with_provider(self, provider):
        """
        Initialization after other components initialized, using Provider class
        instance which is used to return any dependencies (see calculate below).
        """

        self.provider = provider

    def get_requirements(self):
        """
        Return dictionary of derived parameters or other quantities that are needed
        by this component and should be calculated by another theory class.
        """
        return {p: None for p in self.cp_pars}

    def must_provide(self, **requirements):
        """
        Return dictionary of parameters that must be provided.
        """
        return {p: None for p in self.cp_pars}

    def get_can_provide(self):
        """
        Return list of quantities that can be provided.
        """
        return ["Cl", "Dl"]

    def calculate(self, state, want_derived=True, **params_values_dict):
        """
        Calculate the CMB spectra.
        Calls the CosmoPower emulator models.
        """
        # Prepare hand-off to CosmoPower

        pars_for_cp = {p: [state["params"][p]] for p in self.cp_pars}
        if "H0" in self.cp_pars:
            pars_for_cp["h"] = [pars_for_cp["H0"][0] / 100]

        # Get CMB Cls
        state["Cl"] = {
            "ell": self.cp_emulators[list(self.cp_emulators.keys())[0]].modes
        }  # assume all supply the same ell range}
        state["Dl"] = {"ell": state["Cl"]["ell"]}
        for spec_type in list(self.cp_emulators.keys()):
            state["Cl"][spec_type] = (
                self.cp_emulators[spec_type].get_prediction(pars_for_cp).ravel()
            )
            state["Dl"][spec_type] = (
                state["Cl"][spec_type]
                * self.cp_emulators[spec_type].modes
                * (self.cp_emulators[spec_type].modes + 1)
                / (2 * jnp.pi)
            )

    def get_Cl(self, ell_factor=False, **kwargs):
        """Get the Cls or Dls."""
        if ell_factor:
            return self.current_state["Dl"].copy()
        else:
            return self.current_state["Cl"].copy()


# --------------------------------------#
# pycapse wrapper
# --------------------------------------#


class CobayaTheoryPyCapse(cobaya_theory_Theory):
    """
    Wraps a capse model into a cobaya theory class through the pycapse interface.
    See Bonici, Bianchini, Ruiz-Zapatero 2023 for more (https://arxiv.org/abs/2307.14339).

    To initialise pass in the relevant Cobaya block:
    base_path : str
        Base path where emulator files are stored.
    specs_to_emulate : list
        Spectrum type (TT, TE, ...) of corresponding emulator models.

    This code is taken from the Cobaya example for custom theory codes and only slightly modified.
    See https://cobaya.readthedocs.io/en/latest/theories_and_dependencies.html.
    Torrado and Lewis, 2020 (https://arxiv.org/abs/2005.05290)

    Attributes
    ------------
    base_path : str
        Path of the dictionary containing the emulator files.
    specs_to_emulate : list
        List of spectrum types (from TT, TE, EE, pp) indicating which emulators to load.
    pc_emulators : dict
        Pycapse emulators.
    pc_pars : list
        List of parameters required by the emulators. May not match expected parameter order of emulators,
        as it can be the collection of parameters for a series of emulators requiring different inputs.
    pc_pars_to_reg_pars : list
        Dictionary translating the pycapse parameter names to more commonly used ones.
    provider : Provider
        Cobaya provider.
    current_state : dict
        Dict containing current parameters and results.

    Methods
    ---------
    __init__ :
        Initialises an instance of the class.
    initialize :
        Complete set-up.
    initialize_with_provider :
        Initialization after other components initialised.
    get_requirements :
        Returns what the theory code needs to run.
    must_provide :
        Returns what the theory code needs to run.
    get_can_provide :
        Return what the theory code can supply.
    calculate :
        Carry out the calculation.
    get_Cl :
        Return result of calculation (Dl and Cl).
    """

    # I know having these be a class variables is not super elegant, but it's the easiest way to interface with Cobaya
    base_path: str = "./"
    specs_to_emulate: list = []

    def initialize(self):
        """
        Called from __init__ to initialise.
        Loads the emulator models.
        """

        with open(self.base_path + "nn_setup.json") as f:
            nn_setup = json.load(f)

        l = np.load(self.base_path + "l.npy")
        low_crop_ix = np.argwhere(l == 2)
        self.pc_l = l[low_crop_ix:]

        self.pc_emulators = {}
        self.pc_pars = []
        for spec_type in self.specs_to_emulate:
            weights = np.load(self.base_path + f"weights_{spec_type}_lcdm.npy")
            trained_emu = pc.init_emulator(nn_setup, weights, pc.simplechainsemulator)
            emu = pc.cl_emulator(
                trained_emu,
                self.pc_l,
                np.load(self.base_path + "inMinMax_lcdm.npy"),
                np.load(self.base_path + f"outMinMaxCℓ{spec_type}_lcdm.npy"),
            )
            self.pc_emulators[spec_type] = emu
            self.pc_pars += pc.get_parameters_list(emu)
        self.pc_pars = list(np.unique(self.pc_pars))

        self.pc_pars_to_reg_pars = {
            "ln10As": "logA",
            "ns": "ns",
            "H0": "H0",
            "ωb": "ombh2",
            "ωc": "omch2",
            "τ": "tau",
        }

    def initialize_with_provider(self, provider):
        """
        Initialization after other components initialized, using Provider class
        instance which is used to return any dependencies (see calculate below).
        """

        self.provider = provider

    def get_requirements(self):
        """
        Return dictionary of derived parameters or other quantities that are needed
        by this component and should be calculated by another theory class.
        """
        return {p: None for p in self.pc_pars}

    def must_provide(self, **requirements):
        """
        Return dictionary of parameters that must be provided.
        """
        return {p: None for p in self.pc_pars}

    def get_can_provide(self):
        """
        Return list of quantities that can be provided.
        """
        return ["Cl", "Dl"]

    def calculate(self, state, want_derived=True, **params_values_dict):
        """
        Calculate the CMB spectra.
        Hands off to pycapse emulators.
        """
        # Hand-off to pycapse emulators
        state["Cl"] = {"ell": self.pc_l}
        state["Dl"] = {"ell": self.pc_l}
        for spec_type in self.specs_to_emulate:
            par_order = pc.get_parameters_list(self.pc_emulators[spec_type])
            pars_for_pc = np.array(
                [state["params"][self.pc_pars_to_reg_pars[p]] for p in par_order]
            )
            state["Dl"][spec_type] = pc.compute_Cl(
                pars_for_pc, self.pc_emulators[spec_type]
            )
            state["Cl"][spec_type] = (
                pc.compute_Cl(pars_for_pc, self.pc_emulators[spec_type])
                * 2.0
                * np.pi
                / (state["Cl"]["ell"] * (state["Cl"]["ell"] + 1))
            )

    def get_Cl(self, ell_factor=False, **kwargs):
        """Get the Cls or Dls."""
        if ell_factor:
            return self.current_state["Dl"].copy()
        else:
            return self.current_state["Cl"].copy()


# --------------------------------------#
# BB TEMPLATE FOREGROUND - COBAYA INTERFACE
# --------------------------------------#


class CobayaTheoryBBTemplate(cobaya_theory_Theory):
    """
    Wrapper for a two template BB calculator in a cobaya theory class.

    To initialise, pass 'template_filenames' in the relevant cobaya block:
    A dictionary with 'lensing_B_modes' and 'r' template file names.

    This code is taken from the Cobaya example for custom theory codes and only slightly modified.
    See https://cobaya.readthedocs.io/en/latest/theories_and_dependencies.html.
    Torrado and Lewis, 2020 (https://arxiv.org/abs/2005.05290)

    Attributes
    ------------
    template_filenames : dict
        File names of templates.
    provider : Provider
        Cobaya provider.
    current_state : dict
        Dict containing current parameters and results.
    templates : dict
        Dict containing cropped BB spectrum templates.
    ells : array (int)
        Ell range provided.

    Methods
    ---------
    initialize :
        Complete set-up, loading templates.
    initialize_with_provider :
        Initialization after other components initialised.
    get_requirements :
        Returns what the theory code needs to run.
    must_provide :
        Returns what the theory code needs to run.
    get_can_provide :
        Return what the theory code can supply.
    calculate :
        Carry out the calculation.
    get_Cl :
        Return result of calculation.

    """

    # I know having this be a class variable is not super elegant, but it's the easiest way to interface with Cobaya
    template_filenames: dict = {}

    def initialize(self):
        """
        Called from __init__ to initialise.
        Loads the templates requested and crops them to the right ell range.
        """

        # Read in templates
        # Assumes these are CAMB outputs, such that the first column is ell and the fourth one is BB
        # Cast into numpy arrays for now to make ell slicing easier
        r_template_arr = np.array(
            candl.io.read_file_from_path(self.template_filenames["r"]).T
        )
        lensing_B_modes_template_arr = np.array(
            candl.io.read_file_from_path(self.template_filenames["lensing_B_modes"]).T
        )

        # Determine common ell range
        common_ell = np.sort(
            list(
                set(r_template_arr[0, :]).intersection(
                    set(lensing_B_modes_template_arr[0, :])
                )
            )
        )
        smallest_common_ell = np.amin(common_ell)
        largest_common_ell = np.amax(common_ell)
        self.ells = np.arange(smallest_common_ell, largest_common_ell + 1)

        # Crop to common ell range and hold onto templates
        self.templates = {}
        for kw, template_arr in zip(
            ["r", "lensing_B_modes"], [r_template_arr, lensing_B_modes_template_arr]
        ):
            low_ell_ix = list(template_arr[0, :]).index(int(smallest_common_ell))
            high_ell_ix = list(template_arr[0, :]).index(int(largest_common_ell))
            self.templates[kw] = jnp.array(
                template_arr[3, low_ell_ix : high_ell_ix + 1]
            )

    def initialize_with_provider(self, provider):
        """
        Initialization after other components initialized, using Provider class
        instance which is used to return any dependencies (see calculate below).
        """

        self.provider = provider

    def get_requirements(self):
        """
        Return dictionary of derived parameters or other quantities that are needed
        by this component and should be calculated by another theory class.
        """
        return {"r": None, "A_L_BB": None}

    def must_provide(self, **requirements):
        """Return dictionary of parameters that must be provided."""
        return {"r": None, "A_L_BB": None}

    def get_can_provide(self):
        """Return list of quantities that can be provided."""
        return ["Cl"]

    def calculate(self, state, want_derived=True, **params_values_dict):
        """Calculate the CMB spectra. Sums the two templates with their respective amplitudes."""
        # Scale templates and return sum
        BB_spec = (
            state["params"]["r"] * self.templates["r"]
            + state["params"]["A_L_BB"] * self.templates["lensing_B_modes"]
        )
        state["Dl"] = {"BB": BB_spec, "ell": self.ells}
        state["Cl"] = {
            "BB": 2 * jnp.pi * state["Dl"]["BB"] / (self.ells * (self.ells + 1.0)),
            "ell": state["Dl"]["ell"],
        }

    def get_Cl(self, ell_factor=False, **kwargs):
        """Get the Cls or Dls."""
        if ell_factor:
            return self.current_state["Dl"].copy()
        else:
            return self.current_state["Cl"].copy()


# --------------------------------------#
# COSMOPOWER LENSING EMULATOR - COBAYA INTERFACE
# --------------------------------------#


class CobayaTheoryCosmoPowerJAXLensing(cobaya_theory_Theory):
    """
    Wraps CosmoPower-JAX model for lensing into a cobaya theory code.
    See D. Piras, A. Spurio Mancini 2023 and A. Spurio Mancini et al. 2021 for more (https://arxiv.org/abs/2305.06347, https://arxiv.org/abs/2106.03846).

    To initialise, pass 'emulator_filenames' in the relevant cobaya block:
    A dictionary of spectrum types and emulator file names as placed in cosmopower-jax's trained_models/ folder.

    This code is taken from the Cobaya example for custom theory codes and only slightly modified.
    See https://cobaya.readthedocs.io/en/latest/theories_and_dependencies.html.
    Torrado and Lewis, 2020 (https://arxiv.org/abs/2005.05290)

    Attributes
    ------------
    emulator_filenames : dict
        File names of emulators.
    cp_emulators : dict
        CosmoPower emulators.
    cp_pars : list
        List of parameters required by the emulators.
    provider : Provider
        Cobaya provider.
    descriptor : str
        A short descriptor.
    par_names : list
        Names of parameters involved in transformation.
    current_state : dict
        Dict containing current parameters and results.

    Methods
    ---------
    initialize :
        Complete set-up.
    initialize_with_provider :
        Initialization after other components initialised.
    get_requirements :
        Returns what the theory code needs to run.
    must_provide :
        Returns what the theory code needs to run.
    get_can_provide :
        Return what the theory code can supply.
    calculate :
        Carry out the calculation.
    get_Cl :
        Return result of calculation (Cl and Dl).

    """

    # I know having this be a class variable is not super elegant, but it's the easiest way to interface with Cobaya
    emulator_filenames: dict = {}

    def initialize(self):
        """Called from __init__ to initialise. Calls the CosmoPower-JAX emulator models."""

        self.cp_emulators = {}
        for fileType in self.emulator_filenames.keys():
            if fileType == "pp":
                self.cp_emulators[fileType] = CPJ(
                    probe="cmb_pp", filename=self.emulator_filenames[fileType] + ".pkl"
                )
            elif fileType == "TE":
                self.cp_emulators[fileType] = CPJ(
                    probe="custom_pca",
                    filename=self.emulator_filenames[fileType] + ".pkl",
                )
            else:
                self.cp_emulators[fileType] = CPJ(
                    probe="custom_log",
                    filename=self.emulator_filenames[fileType] + ".pkl",
                )

        # changing all of the parameter names
        self.cp_pars = list(self.cp_emulators["pp"].parameters)
        if "h" in self.cp_pars:
            self.cp_pars[self.cp_pars.index("h")] = "H0"
        if "omega_b" in self.cp_pars:
            self.cp_pars[self.cp_pars.index("omega_b")] = "ombh2"
        if "omega_cdm" in self.cp_pars:
            self.cp_pars[self.cp_pars.index("omega_cdm")] = "omch2"
        if "n_s" in self.cp_pars:
            self.cp_pars[self.cp_pars.index("n_s")] = "ns"
        if "ln10^{10}A_s" in self.cp_pars:
            self.cp_pars[self.cp_pars.index("ln10^{10}A_s")] = "logA"
        if "tau_reio" in self.cp_pars:
            self.cp_pars[self.cp_pars.index("tau_reio")] = "tau"

    def initialize_with_provider(self, provider):
        """
        Initialization after other components initialized, using Provider class
        instance which is used to return any dependencies (see calculate below).
        """

        self.provider = provider

    def get_requirements(self):
        """
        Return dictionary of derived parameters or other quantities that are needed
        by this component and should be calculated by another theory class.
        """
        return {p: None for p in self.cp_pars}

    def must_provide(self, **requirements):
        """Return dictionary of parameters that must be provided."""
        return {p: None for p in self.cp_pars}

    def get_can_provide(self):
        """Return list of quantities that can be provided."""
        return ["Cl", "Dl"]

    def calculate(self, state, want_derived=True, **params_values_dict):
        """Calculate the CMB spectra. Calls the CosmoPower-JAX emulator models."""
        # Prepare hand-off to CosmoPower
        # numpy array creation is faster, but need jnp to get derivs
        # This is the order expected by CosmoPower-JAX

        pars_for_cp = [state["params"][p] for p in self.cp_pars]
        if "H0" in self.cp_pars:
            pars_for_cp[self.cp_pars.index("H0")] /= 100
        pars_for_cp = jnp.array(pars_for_cp)

        # Get CMB Cls
        state["Cl"] = {
            "pp": self.cp_emulators["pp"].predict(pars_for_cp).ravel(),
            "L": self.cp_emulators["pp"].modes,
        }
        state["Dl"] = {
            "pp": state["Cl"]["pp"]
            * self.cp_emulators["pp"].modes
            * self.cp_emulators["pp"].modes
            * (self.cp_emulators["pp"].modes + 1)
            * (self.cp_emulators["pp"].modes + 1)
            / (2 * jnp.pi),
            "L": self.cp_emulators["pp"].modes,
        }
        state["Dl"]["kk"] = state["Dl"]["pp"] * (2 * jnp.pi) / 4

        for spec_type in self.cp_emulators.keys():
            if spec_type != "pp":
                state["Cl"][spec_type] = (
                    self.cp_emulators[spec_type].predict(pars_for_cp).ravel()
                )
                state["Dl"][spec_type] = (
                    state["Cl"][spec_type]
                    * self.cp_emulators[spec_type].modes
                    * (self.cp_emulators[spec_type].modes + 1)
                    / (2 * jnp.pi)
                )
                state["Cl"]["ell"] = self.cp_emulators[spec_type].modes
                state["Dl"]["ell"] = self.cp_emulators[spec_type].modes

    def get_Cl(self, ell_factor=False, **kwargs):
        """Get the Cls or Dls."""
        if ell_factor:
            return self.current_state["Dl"].copy()
        else:
            return self.current_state["Cl"].copy()


class CobayaTheoryCosmoPowerLensing(cobaya_theory_Theory):
    """
    Wraps CosmoPower model for lensing into a cobaya theory code.
    See A. Spurio Mancini et al. 2021 for more (https://arxiv.org/abs/2106.03846).

    To initialise, pass 'emulator_filenames' in the relevant cobaya block:
    A dictionary of spectrum types and emulator file names as placed in cosmopower-jax's trained_models/ folder.

    This code is taken from the Cobaya example for custom theory codes and only slightly modified.
    See https://cobaya.readthedocs.io/en/latest/theories_and_dependencies.html.
    Torrado and Lewis, 2020 (https://arxiv.org/abs/2005.05290)

    Attributes
    -------------
    emulator_filenames : dict
        File names of emulators.
    cp_emulators : dict
        CosmoPower emulators.
    cp_pars : list
        List of parameters required by the emulators.
    provider : Provider
        Cobaya provider.

    descriptor : str
        A short descriptor.
    par_names : list
        Names of parameters involved in transformation.
    current_state : dict
        Dict containing current parameters and results.

    Methods
    ---------
    __init__ :
        Initialises an instance of the class.
    initialize :
        Complete set-up.
    initialize_with_provider :
        Initialization after other components initialised.
    get_requirements :
        Returns what the theory code needs to run.
    must_provide :
        Returns what the theory code needs to run.
    get_can_provide :
        Return what the theory code can supply.
    calculate :
        Carry out the calculation.
    get_Cl :
        Return result of calculation (Cls and Dls).
    """

    # I know having this be a class variable is not super elegant, but it's the easiest way to interface with Cobaya
    emulator_filenames: dict = {}

    def initialize(self):
        """Called from __init__ to initialise. Calls the CosmoPower emulator models."""

        self.cp_emulators = {}
        for spec_type in self.emulator_filenames.keys():
            if spec_type == "TE" or spec_type == "pp":
                self.cp_emulators[spec_type] = cp.cosmopower_PCAplusNN(
                    restore=True, restore_filename=self.emulator_filenames[spec_type]
                )
                self.cp_emulators[spec_type].get_prediction = self.cp_emulators[
                    spec_type
                ].predictions_np
            else:
                self.cp_emulators[spec_type] = cp.cosmopower_NN(
                    restore=True, restore_filename=self.emulator_filenames[spec_type]
                )
                self.cp_emulators[spec_type].get_prediction = self.cp_emulators[
                    spec_type
                ].ten_to_predictions_np

        # changing all of the parameter names
        self.cp_pars = list(self.cp_emulators["pp"].parameters)
        if "h" in self.cp_pars:
            self.cp_pars[self.cp_pars.index("h")] = "H0"
        if "omega_b" in self.cp_pars:
            self.cp_pars[self.cp_pars.index("omega_b")] = "ombh2"
        if "omega_cdm" in self.cp_pars:
            self.cp_pars[self.cp_pars.index("omega_cdm")] = "omch2"
        if "n_s" in self.cp_pars:
            self.cp_pars[self.cp_pars.index("n_s")] = "ns"
        if "ln10^{10}A_s" in self.cp_pars:
            self.cp_pars[self.cp_pars.index("ln10^{10}A_s")] = "logA"
        if "tau_reio" in self.cp_pars:
            self.cp_pars[self.cp_pars.index("tau_reio")] = "tau"

    def initialize_with_provider(self, provider):
        """
        Initialization after other components initialized, using Provider class
        instance which is used to return any dependencies (see calculate below).
        """

        self.provider = provider

    def get_requirements(self):
        """
        Return dictionary of derived parameters or other quantities that are needed
        by this component and should be calculated by another theory class.
        """
        return {p: None for p in self.cp_pars}

    def must_provide(self, **requirements):
        """Return dictionary of parameters that must be provided."""
        return {p: None for p in self.cp_pars}

    def get_can_provide(self):
        """Return list of quantities that can be provided."""
        return ["Cl", "Dl"]

    def calculate(self, state, want_derived=True, **params_values_dict):
        """Calculate the CMB spectra. Calls the CosmoPower emulator models."""
        # Prepare hand-off to CosmoPower
        # numpy array creation is faster, but need jnp to get derivs
        # This is the order expected by CosmoPower-JAX

        pars_for_cp = [state["params"][p] for p in self.cp_pars]
        if "H0" in self.cp_pars:
            pars_for_cp[self.cp_pars.index("H0")] /= 100
        pars_for_cp = jnp.array(pars_for_cp)

        # Get CMB Cls and Dls
        state["Cl"] = {
            "pp": self.cp_emulators["pp"].get_predicion(pars_for_cp).ravel(),
            "L": self.cp_emulators["pp"].modes,
        }
        state["Dl"] = {
            "pp": state["Cl"]["pp"]
            * self.cp_emulators["pp"].modes
            * self.cp_emulators["pp"].modes
            * (self.cp_emulators["pp"].modes + 1)
            * (self.cp_emulators["pp"].modes + 1)
            / (2 * jnp.pi),
            "L": state["Cl"]["L"],
        }
        state["Dl"]["kk"] = state["Dl"]["pp"] * (2 * jnp.pi) / 4

        for spec_type in self.cp_emulators.keys():
            if spec_type != "pp":
                state["Cl"][spec_type] = (
                    self.cp_emulators[spec_type].get_prediction(pars_for_cp).ravel()
                )
                state["Dl"][spec_type] = (
                    state["Cl"][spec_type]
                    * self.cp_emulators[spec_type].modes
                    * (self.cp_emulators[spec_type].modes + 1)
                    / (2 * jnp.pi)
                )
                state["Cl"]["ell"] = self.cp_emulators[spec_type].modes
                state["Dl"]["ell"] = state["Cl"]["ell"]

    def get_Cl(self, ell_factor=False, **kwargs):
        """Get the Cls or Dls."""
        if ell_factor:
            return self.current_state["Dl"].copy()
        else:
            return self.current_state["Cl"].copy()


# --------------------------------------#
# LIKELIHOOD - COBAYA INTERFACE
# --------------------------------------#


class CandlCobayaLikelihood(cobaya_likelihood_Likelihood):
    """
    Wrapper for a candl likelihood into a cobaya.likelihood.Likelihood class.
    Based on example likelihood provided by Cobaya (https://cobaya.readthedocs.io/en/latest/likelihoods.html), Torrado and Lewis, 2020 (https://arxiv.org/abs/2005.05290).
    See the different attributes below that can be set when passing the class to Cobaya.

    Attributes
    ------------
    data_set_file : str
        Path of the data set info yaml file.
    variant : str
        Variant of the likelihood (if requested).
    clear_internal_priors : bool
        Whether to clear internal priors.
    clear_specific_priors : str or list
        String of list of strings of specific priors to clear. If a given parameter appears in an internal prior, the full prior will be cleared (beware for multi-dimensional priors). Doesn't do anything if clear_internal_priors is True.
    lensing : bool
        Whether to use the lensing likelihood.
    feedback : bool
        Whether to print feedback when initialising the likelihood.
    data_selection : any
        Data selection to be used. String, list of string, binary mask, or path to a mask are supported.
    candl_like : candl.Like or candl.LensLike
        Candl likelihood.
    wrapper: any
        Set if wanting to access a wrapper likelihood (e.g. clipy).
    additional_args : dict
        Additional arguments to be passed to the likelihood at initialisation.

    Methods
    -----------
    __init__ :
        Initialises an instance of the class.
    initialize :
        Internal set-up (load candl likelihood).
    get_requirements :
        Returns what the likelihood needs to run.
    logp :
        Evaluate the likelihood, calling candl under the hood.

    """

    data_set_file: str = "./"
    variant: str = None
    clear_internal_priors: bool = True
    clear_specific_priors: list = []
    lensing: bool = False
    feedback: bool = True
    data_selection: any = ...
    wrapper: any = None
    additional_args: dict = {}

    def initialize(self):
        """
        Called from __init__ to initialise and complete the setup.
        Loads the candl likelihood.
        """

        # Grab the correct data set
        try:
            # Check whether a short cut from a library of the style module.data_set_name is being passed
            module_name, data_set_name = self.data_set_file.split(".")
            data_module = importlib.import_module(module_name)
            self.data_set_file = getattr(data_module, data_set_name)
        except:
            pass

        # Collect arguments
        init_args = {
            "variant": self.variant,
            "feedback": self.feedback,
        }
        if self.data_selection is not ...:
            init_args["data_selection"] = self.data_selection
        init_args = {
            **self.additional_args,
            **init_args,
        }  # add any additional args to be passed, giving priority to explicitly named arguments

        # Check if a candl wrapper to an external likelihood is requested
        if self.wrapper is not None:
            try:
                # Go through supported wrappers
                if self.wrapper == "clipy":
                    importlib.import_module("clipy")
                    self.candl_like = clipy.clik_candl(
                        self.data_set_file,
                        **self.additional_args,
                    )
                    # Add a list of parameter names that have priors applied
                    self.candl_like.required_prior_parameters = []
                    for p in list(self.candl_like._prior.keys()):
                        if (
                            type(self.candl_like._prior[p]) == list
                            or type(self.candl_like._prior[p]) == tuple
                        ):  # Sometimes clipy returns lists or tuples here
                            for pp in self.candl_like._prior[p]:
                                self.candl_like.required_prior_parameters.append(pp)
                        elif type(self.candl_like._prior[p]) == str:
                            self.candl_like.required_prior_parameters.append(p)
            except:
                raise Exception(
                    f"candl: wrapper likelihood {self.wrapper} could not be initialised!"
                )
            if self.candl_like is None:
                raise Exception(
                    f"candl: wrapper likelihood {self.wrapper} could not be initialised!"
                )
            else:
                return

        # Initialise the candl likelihood
        try:
            if self.lensing:
                self.candl_like = candl.LensLike(
                    self.data_set_file,
                    **init_args,
                )
            else:
                self.candl_like = candl.Like(
                    self.data_set_file,
                    **init_args,
                )
        except:
            raise Exception("candl: likelihood could not be initialised!")

        # by default clear internal priors and assume these are taken care off by Cobaya
        line_width = 80
        if self.clear_internal_priors:
            print("candl: clearing all internal priors.")
            self.candl_like.priors = []
            print(line_width * "-")
        else:
            if len(self.clear_specific_priors) > 0:
                if isinstance(self.clear_specific_priors, str):
                    self.clear_specific_priors = [self.clear_specific_priors]
                new_priors = []
                for ignore_par in self.clear_specific_priors:
                    for i, prior in enumerate(self.candl_like.priors):
                        if ignore_par in prior.par_names:
                            continue
                        else:
                            new_priors.append(prior)
                self.candl_like.priors = new_priors

        # Feedback on priors
        if len(self.candl_like.priors) > 0:
            print(
                f"candl: internal priors will be applied on the following parameters:"
            )
            for i, prior in enumerate(self.candl_like.priors):
                print(", ".join(prior.par_names))
            print(line_width * "-")

    def get_requirements(self):
        """Return dictionary of parameters that are needed"""
        # Cls
        required_pars = {"Cl": {}}
        for spec in self.candl_like.unique_spec_types:
            if spec == "kk":
                required_pars["Cl"][
                    "pp"
                ] = (
                    self.candl_like.ell_max
                )  # Cobaya works with pp for lensing by default rather than kk
            else:
                required_pars["Cl"][spec] = self.candl_like.ell_max

        # Nuisance parameters
        for par in self.candl_like.required_nuisance_parameters:
            required_pars[par] = None

        # Any additional priors
        for par in self.candl_like.required_prior_parameters:
            if not par in list(required_pars.keys()):
                required_pars[par] = None

        return required_pars

    def logp(self, **params):
        """Calculate the log-likelihood by calling candl."""
        # Grab the theory spectra
        Dls = self.provider.get_Cl(
            ell_factor=True, units="muK2"
        )  # grab Dls (ell_factor=True)

        # Cover both lensing conventions if required
        if "pp" in Dls:
            Dls["kk"] = Dls["pp"] * jnp.pi / 2.0

        # Crop spectra to correct ell range
        start_ix = np.argwhere(Dls["ell"] == self.candl_like.ell_min)[0][0]
        stop_ix = np.argwhere(Dls["ell"] == self.candl_like.ell_max)[0][0] + 1

        # Assume spectra handed start at ell=0
        for ky in list(Dls.keys()):
            if ky != "ell":
                if ky in ["pp", "kk"]:
                    Dls[ky] = Dls[ky][start_ix:stop_ix]
                else:
                    Dls[ky.upper()] = Dls[ky][start_ix:stop_ix]
                    if ky.lower() == ky:
                        del Dls[ky]

        pars_to_pass = params
        pars_to_pass["Dl"] = Dls

        # Hand off to the likelihood
        logl = self.candl_like.log_like(pars_to_pass)

        return np.float32(logl)


def get_cobaya_info_dict_for_like(
    like,
    name="candl_like",
    data_selection=...,
    clear_internal_priors=True,
    clear_specific_priors=[],
    feedback=True,
    wrapper=None,
    additional_args={},
):
    """
    Thin wrapper for CandlCobayaLikelihood that returns the class with the requested data set in the format expected by Cobaya.
    Note that since Cobaya prefers to instantiate likelihoods itself, this will instantiate a new instance of the likelihood.
    That means, any modifications to the likelihood object you pass will not be reflected in the likelihood that Cobaya uses.
    Still, the wrapper will try to copy over any data selection - it's worth double checking this with the feedback printout.
    Use the optional keywords to pass arguments to the new Cobaya likelihood object during initialisation.
    Note that candl internal priors are cleared by default, be sure to either apply these in the Cobaya parameter block or override the default behaviour.

    Parameters
    ---------------
    like: candl.Like, candl.LensLike
        A candl likelihood to be used.
    name: str
        Name to give the likelihood in Cobaya
    data_selection : any
        Data selection to be used. None, string, list of string, binary mask, or path to a mask are supported.
    clear_internal_priors : bool
        Whether to clear internal priors.
    clear_specific_priors : str or list
        String or list of strings of specific priors to clear. If a given parameter appears in an internal prior, the full prior will be cleared (beware for multi-dimensional priors). Doesn't do anything if clear_internal_priors is True.
    feedback : bool
        Whether to print feedback when initialising the likelihood.
    wrapper : str
        Identifier of a supported external wrapper likelihood.
    additional_args: dict
        Additional arguments to be passed to the likelihood at initialisation.

    Returns
    ---------------
    dict:
        Dictionary to use in Cobaya's 'likelihood' entry.
    """

    # Throw up a warning about how Cobaya re-instantiates likelihoods
    print(
        "Warning: Cobaya will re-instantiate the likelihood object and earlier modifications (such as manually overwriting attributes, the data model, or the priors) you may have done to your instance will likely not apply unless you have specified them here again."
    )

    # Construct Cobaya dictionary
    cobaya_info = {
        name: {
            "external": CandlCobayaLikelihood,
            "data_set_file": like.data_set_file,
            "clear_internal_priors": clear_internal_priors,
            "clear_specific_priors": clear_specific_priors,
            "feedback": feedback,
            "wrapper": wrapper,
            "additional_args": additional_args,
        }
    }

    # Automatically detect some wrappers
    if like.data_set_file[-5:] == ".clik":
        if cobaya_info[name]["wrapper"] is None:
            cobaya_info[name]["wrapper"] = "clipy"

    # Try to copy over existing data selection
    if isinstance(like, candl.Like) or isinstance(like, candl.LensLike):
        # For native candl likelihoods
        if "data_selection" in like.data_set_dict:
            # Some data selection has already been done
            if data_selection is not ...:
                # Overwrite manually if requested
                print(
                    "Warning: overwriting the data selection that was already set in the likelihood with the one provided to the cobaya dictionary generation function."
                )
                cobaya_info[name]["data_selection"] = data_selection
            else:
                cobaya_info[name]["data_selection"] = like.data_set_dict[
                    "data_selection"
                ]
        else:
            # No data selection in the likelihood, but perhaps here
            if data_selection is not ...:
                cobaya_info[name]["data_selection"] = data_selection
    else:
        # For wrapper likelihoods
        try:
            options = like.init_options
        except:
            options = {}
        cobaya_info[name]["additional_args"] = (
            options | cobaya_info[name]["additional_args"]
        )

    # Add flag for lensing likelihoods
    if isinstance(like, candl.LensLike):
        cobaya_info[name]["lensing"] = True

    return cobaya_info


# --------------------------------------#
# PACKAGE LIKELIHOOD FOR MONTEPYTHON
# --------------------------------------#


def get_montepython_nuisance_param_block_for_like(like):
    """
    Prints out info needed by Montepython .param file for nuisance parameters.

    Parameters
    ---------------
    like: candl.Like
        Likelihood to be used.

    Returns
    ---------------
    None

    """

    for nuisance_par in like.required_nuisance_parameters:
        for prior in like.priors:
            if nuisance_par in prior.par_names:
                prior_central = prior.central_value[prior.par_names.index(nuisance_par)]
                prior_width = np.sqrt(
                    np.diag(prior.prior_covariance)[prior.par_names.index(nuisance_par)]
                )
                mp_str = f"data.parameters['{nuisance_par}'] = [{prior_central}, -1, -1, {prior_width}, 1, 'nuisance']"
                print(mp_str)

    return


# --------------------------------------#
# PAR -> Dls SHORTCUTS FOR VARIOUS THEORY CODES
# --------------------------------------#


def get_CobayaTheory_pars_to_theory_specs_func(theory_calc):
    """
    Helper that returns a simple python function that moves from parameters to spectra using a cobaya.theory.Theory instances.

    Parameters
    ---------------
    theory_calc : cobaya.theory.Theory
        Theory code to calculate theory Dls.

    Returns
    ---------------
    func
        Function that takes a dictionary of parameter values, ell_max, and ell_min (optional) as input and returns a dictionary of CMB spectra (Dl).

    """

    def pars_to_theory_specs(pars, ell_high_cut, ell_low_cut=2):
        # Calculate theory Dls
        new_pars = {"params": deepcopy(pars)}
        theory_calc.calculate(new_pars)

        # Figure out ell range
        N_ell = ell_high_cut - ell_low_cut + 1

        theory_start_ix = (
            np.amax((new_pars["Dl"]["ell"][0], ell_low_cut)) - new_pars["Dl"]["ell"][0]
        )
        theory_stop_ix = (
            np.amin((new_pars["Dl"]["ell"][-1], ell_high_cut))
            + 1
            - new_pars["Dl"]["ell"][0]
        )

        like_start_ix = np.amax((new_pars["Dl"]["ell"][0], ell_low_cut)) - ell_low_cut
        like_stop_ix = (
            np.amin((new_pars["Dl"]["ell"][-1], ell_high_cut)) + 1 - ell_low_cut
        )

        # Slice spectra
        Dls = {"ell": np.arange(ell_low_cut, ell_high_cut + 1)}
        for ky in new_pars["Dl"]:
            if ky != "ell":
                Dls[ky] = jnp.zeros(N_ell)
                Dls[ky] = jax_optional_set_element(
                    Dls[ky],
                    np.arange(like_start_ix, like_stop_ix),
                    new_pars["Dl"][ky][theory_start_ix:theory_stop_ix],
                )
        return Dls

    return pars_to_theory_specs


def get_CosmoPowerJAX_pars_to_theory_specs_func(emulator_filenames):
    """
    Helper that returns a simple python function that moves from parameters to spectra using CosmoPower.
    See D. Piras, A. Spurio Mancini 2023 and A. Spurio Mancini et al. 2021 for more (https://arxiv.org/abs/2305.06347, https://arxiv.org/abs/2106.03846).
    Assumes that all emulators have the same input parameters in the same order.

    Parameters
    ---------------
    emulator_filenames : dict
        Dictionary of spectrum types and emulator file names (if they are placed in cosmopower-jax's trained_models/ folder), or the full path of the files, including in either case the file ending '.pkl' or '.npz'.

    Returns
    ---------------
    func
        Function that takes a dictionary of parameter values, ell_max, and ell_min (optional) as input and returns a dictionary of CMB spectra (Dl).


    """

    cp_emulators = {}
    for spec_type in list(emulator_filenames.keys()):
        if spec_type == "TE":
            try:
                cp_emulators[spec_type] = CPJ(
                    probe="custom_pca", filename=emulator_filenames[spec_type]
                )
            except:
                cp_emulators[spec_type] = CPJ(
                    probe="custom_pca", filepath=emulator_filenames[spec_type]
                )
        else:
            try:
                cp_emulators[spec_type] = CPJ(
                    probe="custom_log", filename=emulator_filenames[spec_type]
                )
            except:
                cp_emulators[spec_type] = CPJ(
                    probe="custom_log", filepath=emulator_filenames[spec_type]
                )

    # Grab input parameter order
    cp_pars = list(list(cp_emulators.values())[0].parameters)
    if "h" in cp_pars:
        cp_pars[cp_pars.index("h")] = "H0"

    def pars_to_theory_specs(pars, ell_high_cut, ell_low_cut=2):
        # Hand-off to CosmoPower-JAX
        pars_for_cp = jnp.zeros(len(cp_pars))
        for p in pars:
            if p in cp_pars:
                pars_for_cp = jax_optional_set_element(
                    pars_for_cp, cp_pars.index(p), pars[p]
                )

        if "H0" in cp_pars:
            pars_for_cp = jax_optional_set_element(
                pars_for_cp, cp_pars.index("H0"), pars["H0"] / 100
            )

        # Get CMB Dls
        Dls = {"ell": np.arange(ell_low_cut, ell_high_cut + 1)}
        for spec_type in list(cp_emulators.keys()):
            # Figure out ell range
            N_ell = ell_high_cut - ell_low_cut + 1

            theory_start_ix = (
                np.amax((cp_emulators[spec_type].modes[0], ell_low_cut))
                - cp_emulators[spec_type].modes[0]
            )
            theory_stop_ix = (
                np.amin((cp_emulators[spec_type].modes[-1], ell_high_cut))
                + 1
                - cp_emulators[spec_type].modes[0]
            )

            like_start_ix = (
                np.amax((cp_emulators[spec_type].modes[0], ell_low_cut)) - ell_low_cut
            )
            like_stop_ix = (
                np.amin((cp_emulators[spec_type].modes[-1], ell_high_cut))
                + 1
                - ell_low_cut
            )

            # Slice spectra
            this_Dl = (
                cp_emulators[spec_type].predict(pars_for_cp).ravel()
                * cp_emulators[spec_type].modes
                * (cp_emulators[spec_type].modes + 1)
                / (2 * jnp.pi)
            )
            Dls[spec_type] = jnp.zeros(N_ell)
            Dls[spec_type] = jax_optional_set_element(
                Dls[spec_type],
                np.arange(like_start_ix, like_stop_ix),
                this_Dl[theory_start_ix:theory_stop_ix],
            )

        return Dls

    return pars_to_theory_specs


def get_CosmoPower_pars_to_theory_specs_func(emulator_filenames):
    """
    Helper that returns a simple python function that moves from parameters to spectra using CosmoPower.
    See A. Spurio Mancini et al. 2021 for more (https://arxiv.org/abs/2106.03846).

    Parameters
    ---------------
    emulator_filenames : dict
        Dictionary of spectrum types and emulator file names.

    Returns
    ---------------
    func
        Function that takes a dictionary of parameter values, ell_max, and ell_min (optional) as input and returns a dictionary of CMB spectra (Dl).

    """

    # Load models and unify prediction methods
    cp_emulators = {}
    for spec_type in list(emulator_filenames.keys()):
        if spec_type == "TE":
            cp_emulators[spec_type] = cp.cosmopower_PCAplusNN(
                restore=True, restore_filename=emulator_filenames[spec_type]
            )
            cp_emulators[spec_type].get_prediction = cp_emulators[
                spec_type
            ].predictions_np
        else:
            cp_emulators[spec_type] = cp.cosmopower_NN(
                restore=True, restore_filename=emulator_filenames[spec_type]
            )
            cp_emulators[spec_type].get_prediction = cp_emulators[
                spec_type
            ].ten_to_predictions_np

    def pars_to_theory_specs(pars, ell_high_cut, ell_low_cut=2):
        # Hand-off to CosmoPower
        pars_for_cp = {str(p): [float(np.atleast_1d(pars[p])[0])] for p in pars}
        if "H0" in pars_for_cp and not "h" in pars_for_cp:
            pars_for_cp["h"] = [pars_for_cp["H0"][0] / 100]

        # Get CMB Dls
        Dls = {"ell": np.arange(ell_low_cut, ell_high_cut + 1)}
        for spec_type in list(cp_emulators.keys()):
            # Figure out ell range
            N_ell = ell_high_cut - ell_low_cut + 1

            theory_start_ix = (
                np.amax((cp_emulators[spec_type].modes[0], ell_low_cut))
                - cp_emulators[spec_type].modes[0]
            )
            theory_stop_ix = (
                np.amin((cp_emulators[spec_type].modes[-1], ell_high_cut))
                + 1
                - cp_emulators[spec_type].modes[0]
            )

            like_start_ix = (
                np.amax((cp_emulators[spec_type].modes[0], ell_low_cut)) - ell_low_cut
            )
            like_stop_ix = (
                np.amin((cp_emulators[spec_type].modes[-1], ell_high_cut))
                + 1
                - ell_low_cut
            )

            # Slice spectra
            this_Dl = (
                cp_emulators[spec_type].get_prediction(pars_for_cp).ravel()
                * cp_emulators[spec_type].modes
                * (cp_emulators[spec_type].modes + 1)
                / (2 * jnp.pi)
            )
            Dls[spec_type] = jnp.zeros(N_ell)
            Dls[spec_type] = jax_optional_set_element(
                Dls[spec_type],
                np.arange(like_start_ix, like_stop_ix),
                this_Dl[theory_start_ix:theory_stop_ix],
            )

        return Dls

    return pars_to_theory_specs


def get_PyCapse_pars_to_theory_specs_func(capse_base_path, specs=["TT", "TE", "EE"]):
    """
    Helper that returns a simple python function that moves from parameters to spectra using PyCapse.
    See Bonici, Bianchini, Ruiz-Zapatero 2023 for more (https://arxiv.org/abs/2307.14339).

    Parameters
    ---------------
    capse_base_path : str
        Path where the PyCapse is located.
    specs : list (optional)
        Which spectra (TT, TE, EE, BB) to try to load.

    Returns
    ---------------
    func
        Function that takes a dictionary of parameter values, ell_max, and ell_min (optional) as input and returns a dictionary of CMB spectra (Dl).

    """

    with open(capse_base_path + "nn_setup.json") as f:
        nn_setup = json.load(f)

    l = np.load(capse_base_path + "l.npy")
    pc_emulators = {}
    for spec_type in specs:
        weights = np.load(capse_base_path + f"weights_{spec_type}_lcdm.npy")
        trained_emu = pc.init_emulator(nn_setup, weights, pc.simplechainsemulator)
        emu = pc.cl_emulator(
            trained_emu,
            l,
            np.load(capse_base_path + "inMinMax_lcdm.npy"),
            np.load(capse_base_path + f"outMinMaxCℓ{spec_type}_lcdm.npy"),
        )
        pc_emulators[spec_type] = emu

    pc_pars_to_reg_pars = {
        "ln10As": "logA",
        "ns": "ns",
        "H0": "H0",
        "ωb": "ombh2",
        "ωc": "omch2",
        "τ": "tau",
    }

    def pars_to_theory_specs(pars, ell_high_cut, ell_low_cut=2):
        # Figure out ell range
        N_ell = ell_high_cut - ell_low_cut + 1

        theory_start_ix = np.amax((l[0], ell_low_cut)) - l[0]
        theory_stop_ix = np.amin((l[-1], ell_high_cut)) + 1 - l[0]

        like_start_ix = np.amax((l[0], ell_low_cut)) - ell_low_cut
        like_stop_ix = np.amin((l[-1], ell_high_cut)) + 1 - ell_low_cut

        # Hand-off to pycapse emulators
        Dls = {"ell": np.arange(ell_low_cut, ell_high_cut + 1)}
        for spec_type in specs:
            par_order = pc.get_parameters_list(pc_emulators[spec_type])
            pars_for_pc = np.array([pars[pc_pars_to_reg_pars[p]] for p in par_order])
            Dls[spec_type] = jnp.zeros(N_ell)
            Dls[spec_type] = jax_optional_set_element(
                Dls[spec_type],
                np.arange(like_start_ix, like_stop_ix),
                pc.compute_Cl(pars_for_pc, pc_emulators[spec_type])[
                    theory_start_ix:theory_stop_ix
                ],
            )

        return Dls

    return pars_to_theory_specs


def get_CAMB_pars_to_theory_specs_func(CAMB_pars):
    """
    Helper that returns a simple python function that moves from parameters to spectra using CAMB.

    Parameters
    ---------------
    CAMB_pars : camb.model.CAMBparams
        CAMBparams for the model. Defines accuracy, ell range, etc.

    Returns
    ---------------
    func
        Function that takes a dictionary of parameter values, ell_max, and ell_min (optional) as input and returns a dictionary of CMB spectra (Dl).

    """

    CAMB_ix = {"TT": 0, "EE": 1, "BB": 2, "TE": 3}

    def pars_to_theory_specs(pars, ell_high_cut, ell_low_cut=2):
        # Set cosmological parameters
        CAMB_pars.set_cosmology(
            H0=pars["H0"],
            ombh2=pars["ombh2"],
            omch2=pars["omch2"],
            mnu=0.06,
            omk=0,
            tau=pars["tau"],
        )
        CAMB_pars.InitPower.set_params(
            As=np.exp(pars["logA"]) * 1e-10, ns=pars["ns"], r=0
        )

        # Calculate spectra
        results = camb.get_results(CAMB_pars)
        powers = results.get_cmb_power_spectra(CAMB_pars, CMB_unit="muK")

        # Figure out ell range
        CAMB_ells = jnp.arange(powers["total"].shape[0])

        # Figure out ell range
        N_ell = ell_high_cut - ell_low_cut + 1

        theory_start_ix = np.amax((CAMB_ells[0], ell_low_cut)) - CAMB_ells[0]
        theory_stop_ix = np.amin((CAMB_ells[-1], ell_high_cut)) + 1 - CAMB_ells[0]

        like_start_ix = np.amax((CAMB_ells[0], ell_low_cut)) - ell_low_cut
        like_stop_ix = np.amin((CAMB_ells[-1], ell_high_cut)) + 1 - ell_low_cut

        # Return as dictionary
        Dls = {
            "ell": np.arange(ell_low_cut, ell_high_cut + 1),
            "pp": powers["lens_potential"][theory_start_ix:theory_stop_ix, 0],
            "kk": powers["lens_potential"][theory_start_ix:theory_stop_ix, 0]
            * jnp.pi
            / 2.0,
        }
        for ky in list(CAMB_ix.keys()):
            Dls[ky] = jnp.zeros(N_ell)
            Dls[ky] = jax_optional_set_element(
                Dls[ky],
                np.arange(like_start_ix, like_stop_ix),
                powers["total"][theory_start_ix:theory_stop_ix, CAMB_ix[ky]],
            )

        return Dls

    return pars_to_theory_specs


def get_CLASS_pars_to_theory_specs_func(CLASS_cosmo):
    """
    Helper that returns a simple python function that moves from parameters to spectra using CLASS.

    Parameters
    ---------------
    CLASS_cosmo : classy.Class
        Class for the model to be evaluated. Need to set desired output options, accuracy settings, ell range, etc.

    Returns
    ---------------
    func
        Function that takes a dictionary of parameter values, ell_max, and ell_min (optional) as input and returns a dictionary of CMB spectra (Dl).

    """

    all_class_pars = [
        p.lower() for p in CLASS_cosmo.__dir__()
    ]  # Grab list of parameters that CLASS understands

    def pars_to_theory_specs(pars, ell_high_cut, ell_low_cut=2):

        # Only pass parameters understood by CLASS
        pars_for_class = {}
        for p in pars:
            if p.lower() in all_class_pars:
                pars_for_class[p] = pars[p]

        # Hand off to CLASS
        CLASS_cosmo.set(pars_for_class)
        CLASS_cosmo.compute()
        class_cls = CLASS_cosmo.lensed_cl(ell_high_cut)
        CLASS_cosmo.struct_cleanup()

        # Figure out ell range
        N_ell = ell_high_cut - ell_low_cut + 1

        theory_start_ix = (
            np.amax((class_cls["ell"][0], ell_low_cut)) - class_cls["ell"][0]
        )
        theory_stop_ix = (
            np.amin((class_cls["ell"][-1], ell_high_cut)) + 1 - class_cls["ell"][0]
        )

        like_start_ix = np.amax((class_cls["ell"][0], ell_low_cut)) - ell_low_cut
        like_stop_ix = np.amin((class_cls["ell"][-1], ell_high_cut)) + 1 - ell_low_cut

        # Return as dictionary of Dls
        Dls = {}
        for ky in list(class_cls.keys()):
            if ky == "ell":
                continue
            if ky == "pp":
                Dls[ky] = (
                    class_cls[ky]
                    * ((class_cls["ell"] * (class_cls["ell"] + 1)) ** 2.0)
                    / (2.0 * jnp.pi)
                )[theory_start_ix:theory_stop_ix]
                Dls["kk"] = Dls["pp"] * jnp.pi / 2.0
                continue
            Dls[ky.upper()] = jnp.zeros(N_ell)
            this_Dls = (
                class_cls[ky]
                * (CLASS_cosmo.T_cmb() ** 2)
                * 1e12
                * class_cls["ell"]
                * (class_cls["ell"] + 1)
                / (2 * np.pi)
            )
            Dls[ky.upper()] = jax_optional_set_element(
                Dls[ky.upper()],
                np.arange(like_start_ix, like_stop_ix),
                this_Dls[theory_start_ix:theory_stop_ix],
            )
        Dls["ell"] = np.arange(ell_low_cut, ell_high_cut + 1)

        return Dls

    return pars_to_theory_specs



"""
IO module that handles reading in data and providing feedback to the user.

Overview:
--------------

* :func:`read_meta_info_from_yaml`
* :func:`load_info_yaml`
* :func:`read_spectrum_info_from_yaml`
* :func:`read_file_from_yaml`
* :func:`read_file_from_path`
* :func:`read_effective_frequencies_from_yaml`
* :func:`read_window_functions_from_yaml`
* :func:`read_transformation_info_from_yaml`
* :func:`read_lensing_M_matrices_from_yaml`
* :func:`like_init_output`
"""

# --------------------------------------#
# IMPORTS
# --------------------------------------#

from candl.lib import *
from astropy.io import fits

# --------------------------------------#
# PROCESS YAML INPUT
# --------------------------------------#


def read_meta_info_from_yaml(data_set_dict):
    """
    Read any meta information.

    Parameters
    --------------
    data_set_dict : dict
        The data set dictionary containing all the information from the input yaml file.

    Returns
    --------------
    str :
        Name of the likelihood.
    """

    return data_set_dict["name"]


def load_info_yaml(yaml_file, variant="default"):
    """
    Read the data set yaml file that contains all the information needed to instantiate the likelihood.
    Navigates through index files if necessary.

    Parameters
    --------------
    yaml_file : str
        Path of the data set yaml file or index file
    variant : str
        Optional, to be used for index files. Indicates the variant of the data set to be read in.

    Returns
    --------------
    dict :
        Data set dictionary.
    str :
        Path of the data set yaml file.
    """

    # Check if this is an index file or a complete data set file already
    dataset_file = yaml_file
    if "index.yaml" in yaml_file:

        # Read index file
        with open(yaml_file, "r") as f:
            index_dict = yaml.load(f, Loader=yaml.loader.SafeLoader)

        # Grab file corresponding to desired variant
        if variant is None:
            variant = "default"
        elif variant not in index_dict:
            print(
                f"Variant '{variant}' not found in index file '{yaml_file}'. Running with default '{index_dict['default']}'."
            )
            variant = "default"

        dataset_file = "/".join(yaml_file.split("/")[:-1]) + "/" + index_dict[variant]

    # Read in the dataset yaml file
    with open(dataset_file, "r") as f:
        data_set_dict = yaml.load(f, Loader=yaml.loader.SafeLoader)

    return data_set_dict, dataset_file


def expand_transformation_block(data_set_dict, block_file):
    """
    Read a .yaml file containing a list of transformation entries.

    Parameters
    --------------
    data_set_dict : dict
        The data set dictionary containing all the information from the input yaml file.
    block_file : str
        Full path of the .yaml block file containing the transformations to be read.

    Returns
    --------------
    list :
        List of transformations form the block file.
    """

    with open(data_set_dict["data_set_path"] + block_file, "r") as f:
        transformation_block = yaml.load(f, Loader=yaml.loader.SafeLoader)

    # Check if any paths are given in here - these should be adjusted so that they are relative to the info .yaml file again
    block_file_path_prefix = "/".join(block_file.split("/")[:-1]) + "/"
    for i, tr in enumerate(transformation_block):
        for tr_attribute in list(tr.keys()):
            if "file" in tr_attribute and "." in tr[tr_attribute]:
                # This is most likely a file path that needs to be adjusted
                transformation_block[i][tr_attribute] = (
                    block_file_path_prefix + tr[tr_attribute]
                )

    return transformation_block


def read_spectrum_info_from_yaml(data_set_dict, lensing=False):
    """
    Read spectrum info.
    If lensing == True only returns first, second, and fourth item from usual returns.

    Parameters
    --------------
    data_set_dict : dict
        The data set dictionary containing all the information from the input yaml file.

    Returns
    --------------
    list :
        List of strings of spectrum identifiers.
    list :
        List of strings of spectrum types.
    list :
        List of lists with two entries giving frequencies for each spectrum.
    int :
        Number of total spectra.
    list :
        List of ints giving number of bins for each spectrum.
    """

    # Extract spectrum order, number of bins, spectrum types, and spectrum frequencies
    spec_order = [list(spec.keys())[0] for spec in data_set_dict["spectra_info"]]
    N_bins = [int(list(spec.values())[0]) for spec in data_set_dict["spectra_info"]]

    # Lensing likelihoods only need some of this info
    if lensing == True:
        return spec_order, spec_order, N_bins

    spec_types = [s[:2] for s in spec_order]
    spec_freqs = [s.split(" ")[1].split("x") for s in spec_order]
    N_spectra_total = len(spec_order)
    N_bins = [int(list(spec.values())[0]) for spec in data_set_dict["spectra_info"]]

    return spec_order, spec_types, spec_freqs, N_spectra_total, N_bins


def read_file_from_yaml(data_set_dict, file_kw):
    """
    Read in a file from the data_set dict.

    Parameters
    --------------
    data_set_dict : dict
        The data set dictionary containing all the information from the input yaml file.
    file_kw : str
        A string that corresponds to a keyword in data_set_dict that specifies the file location relative to
        the base path.

    Returns
    --------------
    array :
        File read into array format.
    """

    # Hand off to file reader
    arr = read_file_from_path(data_set_dict["data_set_path"] + data_set_dict[file_kw])

    return arr


def read_file_from_path(full_path):
    """
    Read in a file (array) from a path.
    Method used to read in band powers and covariance matrix.
    Can read four types:
    (1) Text files. Ending must be ".txt" or ".dat".
    (2) Binary files. These must end in ".bin" and be stored as float64s. Will be turned into a square array if possible.
    (3) Numpy files. These must end in ".npy" and will be read in as a numpy array.
    (4) FITS files. These must end in ".fits".

    Parameters
    --------------
    full_path : str
        The absolute path of the file.

    Returns
    --------------
    array :
        File read into array format.
    """

    arr = None
    file_ending = full_path.split(".")[-1]
    if file_ending == "txt" or file_ending == "dat":
        # plain text file
        arr = jnp.array(np.loadtxt(full_path))
    elif file_ending == "bin":
        # binary file
        arr = jnp.array(np.fromfile(full_path, "float64"))
        square_dim = np.sqrt(float(len(arr)))
        if square_dim.is_integer():
            arr = arr.reshape((int(square_dim), int(square_dim)))
    elif file_ending == "npy":
        # numpy file
        arr = jnp.array(np.load(full_path))
    elif file_ending == "fits":
        with fits.open(full_path) as hdul:
            arr = jnp.array(hdul[1].data)

    return arr


def read_effective_frequencies_from_yaml(data_set_dict):
    """
    Read in effective frequency information from data set dictionary.

    Parameters
    --------------
    data_set_dict : dict
        The data set dictionary containing all the information from the input yaml file.

    Returns
    --------------
    dict :
        Dictionary containing keys for all source types given in the effective frequency yaml file. Each entry is a dictionary with frequency identifiers as keys and effective frequencies as values.
    """

    # Check if effective_frequencies exists (might not be needed for all likelihoods depending on spectra/fg models)
    if not "effective_frequencies" in data_set_dict:
        return None

    # Load effective frequencies, casting frequency id's to strings and effective frequencies to floats
    with open(
        data_set_dict["data_set_path"] + data_set_dict["effective_frequencies"], "r"
    ) as f:
        effective_freq_dict = yaml.load(f, Loader=yaml.loader.SafeLoader)
    effective_frequencies = {}
    for spec_source in list(effective_freq_dict.keys()):
        effective_frequencies[spec_source] = {}
        for ky in list(effective_freq_dict[spec_source].keys()):
            effective_frequencies[spec_source][str(ky)] = float(
                effective_freq_dict[spec_source][ky]
            )

    return effective_frequencies


def read_window_functions_from_yaml(data_set_dict, spec_order, N_bins):
    """
    Read band power window functions using data set dictionary. There are two allowed formats:

    (1) Window functions are saved by spectrum as "{spec}_window_functions.txt"
        The files are arrays of (ell, N_bins+1) size, where the first column gives the theory ell.
    (2) Window functions are saved by bin as "window_{i}.txt" starting at i=0.
        The files are arrays of (ell, N_specs+1) size, where the first column gives the theory ell.
    Generally, the first format is preferred as it allows for spectra of different length.

    Parameters
    --------------
    data_set_dict : dict
        The data set dictionary containing all the information from the input yaml file.
    spec_order : list
        List specifying the order of spectra.
    N_bins : list (int)
        Number of bins for each spectrum.

    Returns
    --------------
    list :
        Band power window functions as a list of N_spectra with (N_ell_theory, N_bins) arrays. Start at ell=2.
    """

    # Check what format the window functions are saved in depending on what files are in the passed folder
    files_in_dir = os.listdir(
        data_set_dict["data_set_path"] + data_set_dict["window_functions_folder"]
    )
    bin_expected_files = [f"window_{i + 1}.txt" for i in range(np.amax(N_bins))]
    spec_expected_files = [
        s.replace(" ", "_") + "_window_functions.txt" for s in spec_order
    ]

    if set(spec_expected_files).issubset(set(files_in_dir)):
        # Format (1): window functions are saved by spectra as "{spec}_window_functions.txt"
        window_functions = []
        for i, spec in enumerate(spec_order):
            this_window = np.loadtxt(
                data_set_dict["data_set_path"]
                + data_set_dict["window_functions_folder"]
                + f"{spec.replace(' ', '_')}_window_functions.txt"
            )
            start_ix = np.argwhere(this_window[:, 0] == 2)[0, 0]  # start at ell = 2
            window_functions.append(this_window[start_ix:, 1:])  # remove ell column

    elif set(bin_expected_files).issubset(set(files_in_dir)):
        # Format (2): window functions are saved as "window_{i}.txt"
        # Read in first window to understand file sizes
        first_window = np.loadtxt(
            data_set_dict["data_set_path"]
            + data_set_dict["window_functions_folder"]
            + "window_1.txt"
        )
        ell_column = first_window[:, 0]
        ell_2_ix = np.argwhere(ell_column == 2)[0, 0]
        N_ell_theory = len(ell_column) - ell_2_ix

        window_functions = [
            np.zeros((N_ell_theory, N_bins[i])) for i in range(len(spec_order))
        ]
        for i in range(np.amax(N_bins)):
            wdw = np.loadtxt(
                data_set_dict["data_set_path"]
                + data_set_dict["window_functions_folder"]
                + f"window_{i + 1}.txt"
            )
            for j, spec in enumerate(spec_order):
                window_functions[j][:, i] = wdw[
                    ell_2_ix:, j + 1
                ]  # starting at ell = 2 and ignore ell column

    else:
        raise Exception("Did not recognise window function format.")
        return None

    return [jnp.array(w) for w in window_functions]


def read_transformation_info_from_yaml(data_set_dict, i_tr):
    """
    Read in information about a specific transformation from data set yaml file.

    Parameters
    --------------
    data_set_dict : dict
        The data set dictionary containing all the information from the input yaml file.
    i_tr : int
        Index of the transformation to be read.

    Returns
    --------------
    str :
        Name of the transformation class.
    dict :
        Dictionary of information passed by the user, required as instantiation Parameters for the specified class.
    """

    tr_name = f"{data_set_dict['data_model'][i_tr]['Module']}"
    tr_passed_args = {
        key: data_set_dict["data_model"][i_tr][key]
        for key in data_set_dict["data_model"][i_tr]
        if key != "Module"
    }

    return tr_name, tr_passed_args


def read_lensing_M_matrices_from_yaml(full_path, Mtype="pp"):
    """
    Read lensing M matrices.
    This function assumes that window functions are saved in a specific format:  window{n}.dat, where n runs over the bin numbers.
    Each file specifies the response function at all L values.
    All bins are read in, any cropping is performed later.

    Parameters
    --------------
    full_path : str
        The absolute path of the folder.
    Mtype : str
        Which M matrices to load, "TT", "TE", "EE", "BB", "pp", or "kk".

    Returns
    --------------
    array :
        M matrices as a (number_of_ells, N_files_total) array.
    """

    # Load in window functions

    # Figure out how many files (i.e. bins) there are to read
    N_files_total = len(
        [
            name
            for name in os.listdir(full_path)
            if name[:7] == "window_" and name[-4:] == ".txt"
        ]
    )

    # find highest L from the first bin's window
    first_window = np.loadtxt(full_path + "window_0.txt")
    last_L = int(first_window[-1, 0])
    number_of_ells = last_L - 1  # starting at ell = 2
    start_ix = np.argwhere(first_window[:, 0] == 2)[0, 0]  # start at ell = 2
    this_window = np.zeros([number_of_ells, N_files_total])

    # Spectrum order
    ix_to_access = {"TT": 0, "TE": 1, "EE": 2, "BB": 3, "pp": 4, "kk": 4}

    # each bin has its own file
    for n in range(N_files_total):
        a_window = np.loadtxt(full_path + "window_" + str(n) + ".txt")
        this_window[:, n] = a_window[
            start_ix:, ix_to_access[Mtype] + 1
        ]  # offset by 1 due to ell column in files

    return this_window


# --------------------------------------#
# FEEDBACK FOR INITIALISATION
# --------------------------------------#


def like_init_output(like):
    """
    Print details after successful initialisation of the likelihood.

    Parameters
    -----------------
    like:
        candl.Like or candl.LensLike

    Returns
    --------------
        None

    """

    # Skip if no feedback is requested
    if "feedback" not in like.data_set_dict:
        return
    if not like.data_set_dict["feedback"]:
        return

    if "log_file" in like.data_set_dict:
        logging.basicConfig(
            filename=like.data_set_dict["like_folder_path"]
            + like.data_set_dict["log_file"],
            encoding="utf-8",
            level=logging.INFO,
            format="%(message)s",
        )
        write_msg = lambda msg: logging.info(msg)
    else:
        write_msg = lambda msg: print(msg)

    line_width = 80

    # Header
    start_str = (
        f"Successfully initialised candl likelihood '{like.data_set_dict['name']}' (type: {type(like)}).\n"
        f"Data loaded from '{like.data_set_dict['data_set_path']}'.\n"
        f"Functional likelihood form: {like.data_set_dict['likelihood_form']}"
    )
    write_msg(start_str)
    write_msg(line_width * "-")

    # Spectrum Order
    spec_str = f"It will analyse the following spectra:\n"
    write_msg(spec_str)
    for i, spec in enumerate(like.spec_order):
        write_msg(
            "{:<12}".format(spec)
            + f" ({like.N_bins[i]} bins, bin centres spanning ell = {sigfig.round(float(like.effective_ells[like.bins_start_ix[i]]), decimals=1)} - {sigfig.round(float(like.effective_ells[like.bins_stop_ix[i]-1]), decimals=1)})"
        )
    write_msg(line_width * "-")

    # Data Model
    if len(like.data_model) == 0:
        dat_model_str = f"No transformation have been initialised in the data model.\n"
    else:
        dat_model_str = f"A data model consisting of {len(like.data_model)} transformations has been initialised.\n"
        dat_model_str += f"The following transformations will be applied to the theory spectra in this order:\n"

    write_msg(dat_model_str)
    for i, tr in enumerate(like.data_model):
        write_msg(f"({i+1}) Name: {tr.descriptor}")
        write_msg(len(str(i)) * " " + f"   Type: {type(tr)}")
    write_msg(line_width * "-")

    # Priors
    if len(like.priors) == 0:
        prior_str = f"No priors will be added to the likelihood in candl internally.\n"
    else:
        prior_str = f"By default, a total of {len(like.priors)} Gaussian priors will be added to the likelihood in candl internally.\n"
        prior_str += "If you are loading the likelihood in sampler wrapper these default priors below will overwritten or replaced according to your selection after initialisation."

    write_msg(prior_str)
    write_msg(line_width * "-")

    return



"""
Common transformations for lensing likelihoods.

Note:
----------------
The transformations of lensing likelihoods act on binned spectra.
This is NOT a comprehensive foreground/data model library. Instead, the classes below are designed
for the data sets implemented in candl and serve as examples that you can use to implement any model you want.

Overview:
----------------

* :class:`BinnedTemplateForeground`
* :class:`ResponseFunctionM`
* :class:`LensingAmplitude`
"""

# --------------------------------------#
# IMPORTS
# --------------------------------------#

from candl.lib import *
import candl.transformations.abstract_base

# --------------------------------------#
# LENSING TRANSFORMATIONS
# --------------------------------------#


class BinnedTemplateForeground(candl.transformations.abstract_base.Foreground):
    """
    Template foreground that is already binned, with a single amplitude parameter.

    Used in the SPT-3G 2018 Lensing likelihood implementation.

    Attributes
    ----------------
    template_arr : array (float)
        Template spectrum and ells.
    template_spec : array (float)
        Template spectrum.
    template_ells : array (int)
        Template ells.
    ells : array (float)
        The ell range the transformation acts on.
    descriptor : str
        A short descriptor.
    par_names : list
        Names of parameters involved in transformation.
    amp_param : str
        The name of the amplitude parameter.
    crop_mask : array (bool)
        Mask of which bins to use.

    Methods
    ---------
    __init__ :
        initialises an instance of the class.
    output :
        gives the additive foreground contribution.
    transform :
        transforms an input spectrum.

    Notes
    ----------------

    User required arguments in data set yaml file:

    * template_file (str) : path to file with template spectrum.
    * amp_param (str) : name of the amplitude parameter.

    Examples
    ----------------

    Example yaml block to use a template foreground: ::

        - Module: "common_lensing.BinnedTemplateForeground"
        template_file: "foreground_templates/spt3g_2018_lensing_foreground_template.txt"
        amp_param: "A_fg"
    """

    def __init__(self, ells, template_arr, amp_param, crop_mask, descriptor=""):
        """
        Initialise a new instance of the BinnedTemplateForeground class.

        Attributes
        ----------------
        ells : array (float)
            The ell range the transformation acts on.
        descriptor : str
            A short descriptor.
        amp_param : str
            The name of the amplitude parameter.
        crop_mask : array (bool)
            Mask of which bins to use.
        template_arr : array (float)
            Array with two columns, the first for ell values, the second for values of the template spectrum.

        Returns
        ----------------
        Foreground
            A new instance of the BinnedTemplateForeground class.
        """

        super().__init__(ells=ells, descriptor=descriptor, param_names=[amp_param])

        # Read in template
        self.template_arr = template_arr
        self.template_spec = self.template_arr[crop_mask, 1]
        self.template_ells = self.template_arr[crop_mask, 0]

        # Grab other args
        self.ells = ells
        self.amp_param = amp_param

    def output(self, sample_params):
        """
        Return foreground spectrum.
        Intended to be overwritten by subclasses.

        Attributes
        ----------------
        sampled_params : dict
            Dictionary of nuisance parameter values.

        Returns
        ----------------
        array, float
            Foreground spectrum.
        """

        return self.template_spec * sample_params[self.amp_param]

    def transform(self, Dls, sample_params):
        """
        Transform spectrum by adding foreground component (result of output method).
        Intended to be overwritten by subclasses.

        Attributes
        ----------------
        Dls : array
            Dls to transform.
        sampled_params : dict
            Dictionary of nuisance parameter values.

        Returns
        ----------------
        array, float
            Transformed spectrum.
        """

        return Dls + self.output(sample_params)


class ResponseFunctionM(candl.transformations.abstract_base.Transformation):
    """
    Calculates :math:`M * (C^{th}) - M * (C^{fid})`.

    Used in the SPT-3G 2018 Lensing and ACT DR6 Lensing likelihood implementations.

    Attributes
    ----------------
    ells : array (float)
        The ell range the transformation acts on.
    descriptor : str
        A short descriptor.
    M_matrix : array (float)
        The matrix of the transformation.
    crop_mask : array (bool)
        Mask of which bins to use.
    operation_hint : str
        Type of the 'transform' operation: 'additive'.

    Methods
    ---------
    __init__ :
        initialises an instance of the class.
    output :
        gives the additive contribution.
    transform :
        Returns a transformed spectrum.

    Notes
    ----------------

    User required arguments in data set yaml file:

    * M_matrices_folder (str) : path to folder with M matrices.
    * Mmodes (list) : list of spectra to use in M matrices.
    * fiducial_correction_file (str) : path to file with the fiducial correction.

    Examples
    ----------------

    Example yaml block: ::

        - Module: "common_lensing.ResponseFunctionM"
          M_matrices_folder: "lens_delta_windows_phionly/"
          Mmodes:
            - pp
            - TT
          fiducial_correction_file: "spt3g_2018_pp_lensing_fiducial_correction_phionly.txt"
    """

    def __init__(
        self,
        ells,
        M_matrices,
        fiducial_correction,
        crop_mask,
        descriptor="ResponseFunctionM",
    ):
        """
        Initialise the ResponseFunctionM transformation.

        Attributes
        ----------------
        ells : array (float)
            The ell range the transformation acts on.
        descriptor : str
            A short descriptor.
        M_matrices : dictionary of array (float)
            A dictionary with the matrix(es) of the transformation.
        fiducial_correction : array (float)
            An array with the fiducial correction (M * C_fid) values.
        crop_mask : array (bool)
            Mask of which bins to use.

        Returns
        ----------------
        Transformation
            A new instance of the ResponseFunctionM class.
        """

        super().__init__(ells=ells, descriptor=descriptor, operation_hint="additive")

        # Crop down fiducial correction and M_matrices for bin selection
        self.M_matrices = {}
        for mode in list(M_matrices.keys()):
            self.M_matrices[mode] = jnp.array(M_matrices[mode][:, crop_mask])
        self.fiducial_correction = jnp.array(fiducial_correction[crop_mask])

    @partial(jit, static_argnums=(0,))
    def output(self, sample_params):
        """
        Return the correction.
        Needs to use unbinned (from params) spectra.

        Attributes
        ----------------
        sample_params dict
            Contains the unbinned (pp, TT) theory spectra to be used in calculating (M * C).

        Returns
        ----------------
        array, float
            Response function contribution.
        """

        # M * (Cth) - M * (Cfid)

        # get required modes and fiducial correction
        M_modes = self.M_matrices.keys()
        M_correction = -self.fiducial_correction

        # multiply arrays according to length of theory spectra
        for mode in M_modes:
            M_correction += jnp.dot(
                np.transpose(self.M_matrices[mode][: len(self.ells)]),
                jnp.block(sample_params["Dl"][mode]),
            )
        return M_correction

    @partial(jit, static_argnums=(0,))
    def transform(self, Dls, sample_params):
        """
        Transform the input spectrum.

        Attributes
        ----------------
        Dictionary of Dls : dict of array (float)
            The binned spectra (pp) to add to (M * C).
        sample_params dict
            Contains the unbinned (pp, TT) theory spectra to be used in calculating (M * C).

        Returns
        ----------------
        array : float
            Response function contribution.
        """

        return Dls + self.output(sample_params)


class LensingAmplitude(candl.transformations.abstract_base.Transformation):
    """
    Lensing amplitude function, which multiplies the input by some factor.

    Notes
    ----------------

    User required arguments in data set yaml file:

    * amp_param (str) : name of the amplitude parameter.

    Examples
    ----------------

    Example yaml block to a template foreground: ::

        - Module: "common_lensing.LensingAmplitude"
          amp_param: "AL"

    Methods
    ---------
    __init__ :
        initialises an instance of the class.
    output :
        amp_param.
    transform :
        transforms an input spectrum as amp_param * input.

    Attributes
    ----------------
    ells : array (float)
        The ell range the transformation acts on.
    descriptor : str
        A short descriptor.
    par_names : list
        Names of parameters involved in transformation.
    amp_param : str
        The name of the amplitude parameter.
    operation_hint : str
        Type of the 'transform' operation: 'additive'.
    """

    def __init__(self, ells, amp_param, descriptor=""):
        """
        Initialise a new instance of the LensingAmplitude class.

        Attributes
        ----------------
        ells : array (float)
            The ell range the transformation acts on.
        descriptor : str
            A short descriptor.
        amp_param : str
            The name of the amplitude parameter.

        Returns
        ----------------
        Foreground
            A new instance of the LensingAmplitude class.
        """

        super().__init__(
            ells=ells,
            descriptor=descriptor,
            param_names=[amp_param],
            operation_hint="multiplicative",
        )

        # Grab other args
        self.ells = ells
        self.amp_param = amp_param

    def output(self, sample_params):
        """
        Return multiplicative factor.
        Intended to be overwritten by subclasses.

        Attributes
        ----------------
        sampled_params : dict
            Dictionary of nuisance parameter values.

        Returns
        ----------------
        float
            Multiplicative factor.
        """

        return sample_params[self.amp_param]

    def transform(self, Dls, sample_params):
        """
        Transform spectrum by multiplying by overall constant factor (result of output method).
        Intended to be overwritten by subclasses.

        Attributes
        ----------------
        Dls : array
            Dls to transform.
        sampled_params : dict
            Dictionary of nuisance parameter values.

        Returns
        ----------------
        array, float
            Transformed spectrum.
        """

        return Dls * self.output(sample_params)


class NormCorr(candl.transformations.abstract_base.Transformation):
    """
    Calculates :math:`-2 \\frac{\\frac{d A}{d C} \Delta C}{f} C^{fid}`.

    Used in the ACT DR6 Lensing likelihood.

    Attributes
    ----------------
    ells : array (float)
        The ell range the transformation acts on.
    descriptor : str
        A short descriptor.
    cl2dl : array (float)
        Conversion helper.
    dAL_dC : array (bool)
        Array for matrix multiplication (see equation in docs).
    fAl : array (bool)
        Array for normalisation (see equation in docs).
    template_arr_fiducial_Cl_kk : array (bool)
        Fiducial lensing spectrum.
    fiducial_correction : array (bool)
        Fiducial correction.
    window_functions : list
        List of window functions to use for binning.
    operation_hint : str
        Type of the 'transform' operation: 'additive'.

    Methods
    ---------
    __init__ :
        initialises an instance of the class.
    output :
        gives the additive contribution.
    transform :
        Returns a transformed spectrum.

    Notes
    ----------------

    User required arguments in data set yaml file:

    * M_matrices_folder (str) : path to folder with M matrices.
    * Mmodes (list) : list of spectra to use in M matrices.
    * fiducial_correction_file (str) : path to file with the fiducial correction.

    Examples
    ----------------

    Example yaml block: ::

        - Module: "common_lensing.NormCorr"
          template_file_dAL_dC_TT: "data/ACT_DR6_Lens_v1/dAL_dC_TT.txt"
          template_file_dAL_dC_TE: "data/ACT_DR6_Lens_v1/dAL_dC_TE.txt"
          template_file_dAL_dC_EE: "data/ACT_DR6_Lens_v1/dAL_dC_EE.txt"
          template_file_dAL_dC_BB: "data/ACT_DR6_Lens_v1/dAL_dC_BB.txt"
          template_file_fAL: "data/ACT_DR6_Lens_v1/fAL.txt"
          template_file_fiducial_correction: "data/ACT_DR6_Lens_v1/norm_corr_fiducial_correction.txt"
          template_file_fiducial_Cl_kk: "data/ACT_DR6_Lens_v1/fiducial_Cl_kk.txt"

    """

    def __init__(
        self,
        ells,
        template_arr_dAL_dC_TT,
        template_arr_dAL_dC_TE,
        template_arr_dAL_dC_EE,
        template_arr_dAL_dC_BB,
        template_arr_fAL,
        template_arr_fiducial_Cl_kk,
        template_arr_fiducial_correction,
        window_functions,
        descriptor="NormCorr",
    ):
        """
        Initialise a new instance of the LensingAmplitude class.

        Attributes
        ----------------
        ells : array (float)
            The ell range the transformation acts on.
        descriptor : str
            A short descriptor.
        amp_param : str
            The name of the amplitude parameter.

        Returns
        ----------------
        Foreground
            A new instance of the LensingAmplitude class.
        """

        super().__init__(
            ells=ells,
            descriptor=descriptor,
            operation_hint="additive",
        )

        # Grab other args
        self.ells = ells
        self.cl2dl = ells * (ells + 1) / (2.0 * np.pi)
        self.dAL_dC = {
            "TT": template_arr_dAL_dC_TT,
            "TE": template_arr_dAL_dC_TE,
            "EE": template_arr_dAL_dC_EE,
            "BB": template_arr_dAL_dC_BB,
        }
        self.fAL = template_arr_fAL
        self.template_arr_fiducial_Cl_kk = template_arr_fiducial_Cl_kk
        self.fiducial_correction = template_arr_fiducial_correction
        self.window_functions = window_functions

    @partial(jit, static_argnums=(0,))
    def output(self, sample_params):
        """ """

        # multiply arrays according to length of theory spectra
        norm_corr = 0
        for i, spec in enumerate(["TT", "TE", "EE", "BB"]):
            norm_corr += -2.0 * (
                self.dAL_dC[spec] @ (sample_params["Dl"][spec] / self.cl2dl)
            )
        norm_corr = norm_corr / self.fAL
        norm_corr = norm_corr * self.template_arr_fiducial_Cl_kk

        # now bin
        binned_norm_corr = self.window_functions[0].T @ norm_corr
        binned_norm_corr -= self.fiducial_correction

        return binned_norm_corr

    @partial(jit, static_argnums=(0,))
    def transform(self, Dls, sample_params):
        return Dls + self.output(sample_params)



"""Differentiable Likelihood for CMB Analysis"""

__author__ = ["L. Balkenhol", "C. Trendafilova", "K. Benabed", "S. Galli"]
__version__ = "2.0.3"
__description__ = "Differentiable Likelihood for CMB Analysis."

from .likelihood import *



"""
Galactic Foreground transformations with band pass integrals intended for BB analysis.

Foreground Classes:
--------
GalacticDustBandPass
SynchrotronBandPass
GalacticDustSynchrotronCorrelationBandPass
GalacticDustFixedAlphaBandPass
SynchrotronFixedAlphaBandPass
GalacticDustSynchrotronCorrelationFixedAlphaBandPass
"""

# --------------------------------------#
# IMPORTS
# --------------------------------------#

from candl.lib import *
import candl.transformations.abstract_base
import candl.transformations.common

# --------------------------------------#
# DUST AND SYNCHROTRON WITH BAND PASS INTEGRALS
# --------------------------------------#


class GalacticDustBandPass(candl.transformations.abstract_base.ForegroundBandPass):
    """
    Dusty foreground with modified black-body frequency scaling with integral over band pass with a power law ell
    power spectrum.
    $ A * f(beta, bdp_1) * f(beta, bdp_2) * \left( \ell / \ell_{ref} \right)^(\alpha + 2) $
    where:
        A : amplitude
        ell_{ref} : reference ell
        beta : frequency scaling parameter
        alpha : power law index
        f(beta, bdp_1) : is the frequency scaling given the band pass
    The +2 to the exponent is convention.

    User required arguments in data set yaml file
    ---------
    ell_ref : float
        Reference ell.
    nu_ref : float
        Reference frequency.
    T_GALDUST : float
        Temperature of the dust.
    amp_param : str
        The name of the amplitude parameter.
    alpha_param : str
        The name of the power law index parameter.
    beta_param : str
        The name of the frequency scaling parameter.
    affected_specs : str
        List of spectrum identifiers the transformation is applied to.

    Example yaml block to add polarised galactic dust to all EE spectra:
    - Module: "candl.transformations.common.GalacticDustBandPass"
      amp_param: "BB_GalDust_BDP_Amp"
      alpha_param: "BB_GalDust_BDP_Alpha"
      beta_param: "BB_GalDust_BDP_Beta"
      nu_ref: 353
      affected_specs: ["BB 90x90", "BB 90x150", "BB 90x220", "BB 150x150", "BB 150x220", "BB 220x220"]
      ell_ref: 80
      T_GALDUST: 19.6
      descriptor: "BB Polarised Galactic Dust (Bandpass)"

    Methods
    ---------
    __init__ :
        initialises an instance of the class.
    output :
        gives the additive foreground contribution.
    transform :
        transforms an input spectrum.

    Attributes
    -------
    ells : array (float)
        The ell range the transformation acts on.
    descriptor : str
        A short descriptor.
    par_names : list
        Names of parameters involved in transformation.
    spec_order : array (str)
        Identifiers of spectra in the order in which spectra are handled in the long data vector.
    bandpass_info : list
        List of lists, where each sublist contains the two candl.transformations.abstract_base.BandPass instances for the two
        frequencies involved.
    ell_ref : int
        Reference ell for normalisation.
    nu_ref : float
        Reference frequency.
    N_spec : int
        The total number of spectra in the long data vector.
    affected_specs : list (str)
        List of the spectra to apply this foreground to.
    spec_mask : array (int)
        Masks which spectra of the long data vector are affected by the transformation.
    full_mask : array (int)
        Masks which elements of the long data vector are affected by the transformation.
    T_dust : float
        Dust temperature.
    amp_param : str
        The name of the amplitude parameter.
    alpha_param : str
        The name of the power law index parameter.
    beta_param : str
        The name of the frequency scaling parameter.

    See also
    -------
    dust_frequency_scaling_bandpass
    """

    def __init__(
        self,
        ells,
        spec_order,
        bandpass_info,
        affected_specs,
        amp_param,
        alpha_param,
        beta_param,
        ell_ref,
        nu_ref,
        T_GALDUST,
        descriptor="Galactic Dust (Band pass)",
    ):
        """
        Initialise a new instance of the GalacticDustBandPass class.

        Arguments
        -------
        ells : array (float)
            The ell range the transformation acts on.
        descriptor : str
            A short descriptor.
        bandpass_info : list
            List of lists, where each sublist contains the two candl.transformations.abstract_base.BandPass instances for the two
            frequencies involved.
        spec_order : array (str)
            Identifiers of spectra in the order in which spectra are handled in the long data vector.
        affected_specs : list (str)
            List of the spectra to apply this foreground to.
        ell_ref : int
            Reference ell for normalisation.
        nu_ref : float
            Reference frequency.
        T_GALDUST : float
            Temperature of the CIB.
        amp_param : str
            The name of the amplitude parameter.
        alpha_param : str
            The name of the power law index parameter.
        beta_param : str
            The name of the frequency scaling parameter.

        Returns
        -------
        GalacticDustBandPass
            A new instance of the class.
        """

        super().__init__(
            ells=ells,
            spec_order=spec_order,
            bandpass_info=bandpass_info,
            ell_ref=ell_ref,
            nu_ref=nu_ref,
            descriptor=descriptor,
            param_names=[amp_param, alpha_param, beta_param],
        )

        self.T_dust = T_GALDUST
        self.amp_param = amp_param
        self.alpha_param = alpha_param
        self.beta_param = beta_param

        self.affected_specs = affected_specs
        self.spec_mask = jnp.asarray(
            [spec in self.affected_specs for spec in self.spec_order]
        )

        # Turn spectrum mask into a full mask
        self.full_mask = jnp.asarray(
            jnp.repeat(self.spec_mask, len(self.ells)), dtype=float
        )

    @partial(jit, static_argnums=(0,))
    def output(self, sample_params):
        """
        Return foreground spectrum.

        Arguments
        -------
        sampled_params : dict
            Dictionary of nuisance parameter values.

        Returns
        -------
        array, float
            Foreground spectrum.
        """

        # amplitude part (frequency scaling)
        amp_vals = jnp.array(
            [
                candl.transformations.common.dust_frequency_scaling_bandpass(
                    sample_params[self.beta_param],
                    self.T_dust,
                    self.nu_ref,
                    self.bandpass_info[i][0].nu_spacing,
                    self.bandpass_info[i][0].nu_vals,
                    self.bandpass_info[i][0].bandpass_vals,
                    self.bandpass_info[i][0].thermo_conv[self.nu_ref],
                )
                * candl.transformations.common.dust_frequency_scaling_bandpass(
                    sample_params[self.beta_param],
                    self.T_dust,
                    self.nu_ref,
                    self.bandpass_info[i][1].nu_spacing,
                    self.bandpass_info[i][1].nu_vals,
                    self.bandpass_info[i][1].bandpass_vals,
                    self.bandpass_info[i][1].thermo_conv[self.nu_ref],
                )
                for i in range(self.N_spec)
            ]
        )

        amp_vals *= sample_params[self.amp_param]
        tiled_amp_vals = jnp.repeat(amp_vals, len(self.ells))

        # ell part
        ell_dependence = (self.ells / self.ell_ref) ** (
            sample_params[self.alpha_param] + 2
        )
        tiled_ell_dependence = jnp.tile(
            ell_dependence, self.N_spec
        )  # tiled ell dependence

        # Complete foreground contribution and mask down
        fg_pow = self.full_mask * tiled_amp_vals * tiled_ell_dependence
        return fg_pow

    @partial(jit, static_argnums=(0,))
    def transform(self, Dls, sample_params):
        """
        Transform spectrum by adding foreground component (result of output method).

        Arguments
        -------
        Dls : array
            Dls to transform.
        sampled_params : dict
            Dictionary of nuisance parameter values.

        Returns
        -------
        array, float
            Transformed spectrum.
        """

        return Dls + self.output(sample_params)


class SynchrotronBandPass(candl.transformations.abstract_base.ForegroundBandPass):
    """
    Synchrotron foreground with modified power law frequency scaling with integral over band pass with a power law ell
    power spectrum.
    $ A * f(beta, bdp_1) * f(beta, bdp_2) * \left( \ell / \ell_{ref} \right)^(\alpha) $
    where:
        A : amplitude
        ell_{ref} : reference ell
        beta : frequency scaling parameter
        alpha_param : power law index
        f(beta, bdp_1) : is the frequency scaling given the band pass

    User required arguments in data set yaml file
    ---------
    ell_ref : float
        Reference ell.
    nu_ref : float
        Reference frequency.
    amp_param : str
        The name of the amplitude parameter.
    alpha_param : str
        The name of the power law index parameter.
    beta_param : str
        The name of the frequency scaling parameter.
    affected_specs : str
        List of spectrum identifiers the transformation is applied to.

    Example yaml block to add synchrotron power to all BB spectra:
    - Module: "candl.transformations.common.SynchrotronBandPass"
      amp_param: "BB_Sync_BDP_Amp"
      alpha_param : "BB_Sunc_BDP_Alpha"
      beta_param: "BB_Sync_BDP_Beta"
      nu_ref: 150
      affected_specs: ["BB 90x90", "BB 90x150", "BB 90x220", "BB 150x150", "BB 150x220", "BB 220x220"]
      ell_ref: 80
      descriptor: "BB Synchrotron (Bandpass)"

    Methods
    ---------
    __init__ :
        initialises an instance of the class.
    output :
        gives the additive foreground contribution.
    transform :
        transforms an input spectrum.

    Attributes
    -------
    ells : array (float)
        The ell range the transformation acts on.
    descriptor : str
        A short descriptor.
    par_names : list
        Names of parameters involved in transformation.
    spec_order : array (str)
        Identifiers of spectra in the order in which spectra are handled in the long data vector.
    bandpass_info : list
        List of lists, where each sublist contains the two candl.transformations.abstract_base.BandPass instances for the two
        frequencies involved.
    ell_ref : int
        Reference ell for normalisation.
    nu_ref : float
        Reference frequency.
    N_spec : int
        The total number of spectra in the long data vector.
    affected_specs : list (str)
        List of the spectra to apply this foreground to.
    spec_mask : array (int)
        Masks which spectra of the long data vector are affected by the transformation.
    full_mask : array (int)
        Masks which elements of the long data vector are affected by the transformation.
    amp_param : str
        The name of the amplitude parameter.
    alpha_param : str
        The name of the power law index parameter.
    beta_param : str
        The name of the frequency scaling parameter.

    See also
    -------
    sync_frequency_scaling_bandpass
    """

    def __init__(
        self,
        ells,
        spec_order,
        bandpass_info,
        affected_specs,
        amp_param,
        alpha_param,
        beta_param,
        ell_ref,
        nu_ref,
        descriptor="Synchrotron (Band pass)",
    ):
        """
        Initialise a new instance of the SynchrotronBandPass class.

        Arguments
        -------
        ells : array (float)
            The ell range the transformation acts on.
        descriptor : str
            A short descriptor.
        bandpass_info : list
            List of lists, where each sublist contains the two candl.transformations.abstract_base.BandPass instances for the two
            frequencies involved.
        spec_order : array (str)
            Identifiers of spectra in the order in which spectra are handled in the long data vector.
        affected_specs : list (str)
            List of the spectra to apply this foreground to.
        ell_ref : int
            Reference ell for normalisation.
        nu_ref : float
            Reference frequency.
        amp_param : str
            The name of the amplitude parameter.
        alpha_param : str
            The name of the power law index parameter.
        beta_param : str
            The name of the frequency scaling parameter.

        Returns
        -------
        SynchrotronBandPass
            A new instance of the class.
        """

        super().__init__(
            ells=ells,
            spec_order=spec_order,
            bandpass_info=bandpass_info,
            ell_ref=ell_ref,
            nu_ref=nu_ref,
            descriptor=descriptor,
            param_names=[amp_param, alpha_param, beta_param],
        )

        self.amp_param = amp_param
        self.beta_param = beta_param
        self.alpha_param = alpha_param

        self.affected_specs = affected_specs
        self.spec_mask = jnp.asarray(
            [spec in self.affected_specs for spec in self.spec_order]
        )

        # Turn spectrum mask into a full mask
        self.full_mask = jnp.asarray(
            jnp.repeat(self.spec_mask, len(self.ells)), dtype=float
        )

    @partial(jit, static_argnums=(0,))
    def output(self, sample_params):
        """
        Return foreground spectrum.

        Arguments
        -------
        sampled_params : dict
            Dictionary of nuisance parameter values.

        Returns
        -------
        array, float
            Foreground spectrum.
        """

        # amplitude part (frequency scaling)
        amp_vals = jnp.array(
            [
                candl.transformations.common.sync_frequency_scaling_bandpass(
                    sample_params[self.beta_param],
                    self.nu_ref,
                    self.bandpass_info[i][0].nu_spacing,
                    self.bandpass_info[i][0].nu_vals,
                    self.bandpass_info[i][0].bandpass_vals,
                    self.bandpass_info[i][0].thermo_conv[self.nu_ref],
                )
                * candl.transformations.common.sync_frequency_scaling_bandpass(
                    sample_params[self.beta_param],
                    self.nu_ref,
                    self.bandpass_info[i][1].nu_spacing,
                    self.bandpass_info[i][1].nu_vals,
                    self.bandpass_info[i][1].bandpass_vals,
                    self.bandpass_info[i][1].thermo_conv[self.nu_ref],
                )
                for i in range(self.N_spec)
            ]
        )

        amp_vals *= sample_params[self.amp_param]
        tiled_amp_vals = jnp.repeat(amp_vals, len(self.ells))

        # ell part
        ell_dependence = (self.ells / self.ell_ref) ** (sample_params[self.alpha_param])
        tiled_ell_dependence = jnp.tile(
            ell_dependence, self.N_spec
        )  # tiled ell dependence

        # Complete foreground contribution and mask down
        fg_pow = self.full_mask * tiled_amp_vals * tiled_ell_dependence
        return fg_pow

    @partial(jit, static_argnums=(0,))
    def transform(self, Dls, sample_params):
        """
        Transform spectrum by adding foreground component (result of output method).

        Arguments
        -------
        Dls : array
            Dls to transform.
        sampled_params : dict
            Dictionary of nuisance parameter values.

        Returns
        -------
        array, float
            Transformed spectrum.
        """

        return Dls + self.output(sample_params)


# --------------------------------------#
# DUST AND SYNCHROTRON WITH BAND PASS INTEGRALS - FIXED ALPHA
# --------------------------------------#


class GalacticDustSynchrotronCorrelationBandPass(
    candl.transformations.abstract_base.ForegroundBandPass
):
    """
    Correlation between galactic dust and synchrotron.
    Uses geometric mean of GalacticDustBandPass and SynchrotronBandPass, intended to be used in conjunction with these
    two classes.
    Note: assumes that both have the same ell_ref!

    User required arguments in data set yaml file
    ---------
    ell_ref : float
        Reference ell.
    nu_ref_dust : float
        Dust reference frequency.
    nu_ref_sync : float
        Synchrotron reference frequency.
    correlation_param : str
        Name of the correlation parameter.
    T_GALDUST : float
        Temperature of the dust.
    dust_amp_param : str
        The name of the dust amplitude parameter.
    dust_alpha_param : str
        The name of the dust power law index parameter.
    dust_beta_param : str
        The name of the dust frequency scaling parameter.
    affected_specs : str
        List of spectrum identifiers the transformation is applied to.
    sync_amp_param : str
        The name of the synchrotron amplitude parameter.
    sync_alpha_param : str
        The name of the synchrotron power law index parameter.
    sync_beta_param : str
        The name of the synchrotron frequency scaling parameter.


    Example yaml block to add correlation to BB spectra:
    - Module: "candl.transformations.low_ell_BB.GalacticDustSynchrotronCorrelationBandPass"
      correlation_param: "BB_rho_dust_sync"
      dust_amp_param: "BB_GalDust_BDP_Amp"
      dust_alpha_param: "BB_GalDust_BDP_Alpha"
      dust_beta_param: "BB_GalDust_BDP_Beta"
      sync_amp_param: "BB_Sync_BDP_Amp"
      sync_alpha_param: "BB_Sync_BDP_Alpha"
      sync_beta_param: "BB_Sync_BDP_Beta"
      nu_ref_dust: 353
      nu_ref_sync: 23
      affected_specs: ["BB 90x90", "BB 90x150", "BB 150x150"]
      ell_ref: 80
      T_GALDUST: 19.6
      descriptor: "BB Dust Synchrotron Correlation (Bandpass)"

    Methods
    ---------
    __init__ :
        initialises an instance of the class.
    output :
        gives the additive foreground contribution.
    transform :
        transforms an input spectrum.

    Attributes
    -------
    ells : array (float)
        The ell range the transformation acts on.
    descriptor : str
        A short descriptor.
    par_names : list
        Names of parameters involved in transformation.
    spec_order : array (str)
        Identifiers of spectra in the order in which spectra are handled in the long data vector.
    bandpass_info : list
        List of lists, where each sublist contains the two candl.transformations.abstract_base.BandPass instances for the two
        frequencies involved.
    ell_ref : int
        Reference ell for normalisation.
    nu_ref_dust : float
        Dust reference frequency.
    nu_ref_sync : float
        Synchrotron reference frequency.
    N_spec : int
        The total number of spectra in the long data vector.
    affected_specs : list (str)
        List of the spectra to apply this foreground to.
    spec_mask : array (int)
        Masks which spectra of the long data vector are affected by the transformation.
    full_mask : array (int)
        Masks which elements of the long data vector are affected by the transformation.
    T_dust : float
        Dust temperature.
    dust_amp_param : str
        The name of the dust amplitude parameter.
    dust_alpha_param : str
        The name of the dust power law index parameter.
    dust_beta_param : str
        The name of the dust frequency scaling parameter.
    sync_amp_param : str
        The name of the sync amplitude parameter.
    sync_alpha_param : str
        The name of the sync power law index parameter.
    sync_beta_param : str
        The name of the sync frequency scaling parameter.
    correlation_param : str
        Name of the correlation parameter.

    See also
    -------
    dust_frequency_scaling_bandpass
    """

    def __init__(
        self,
        ells,
        spec_order,
        bandpass_info,
        affected_specs,
        correlation_param,
        dust_amp_param,
        dust_alpha_param,
        dust_beta_param,
        sync_amp_param,
        sync_alpha_param,
        sync_beta_param,
        ell_ref,
        nu_ref_dust,
        nu_ref_sync,
        T_GALDUST,
        descriptor="Dust Synchrotron Correlation (Band pass)",
    ):
        """
        Initialise a new instance of the GalacticDustSynchrotronCorrelationBandPass class.

        Arguments
        -------
        ells : array (float)
            The ell range the transformation acts on.
        descriptor : str
            A short descriptor.
        bandpass_info : list
            List of lists, where each sublist contains the two candl.transformations.abstract_base.BandPass instances for the two
            frequencies involved.
        spec_order : array (str)
            Identifiers of spectra in the order in which spectra are handled in the long data vector.
        affected_specs : list (str)
            List of the spectra to apply this foreground to.
        ell_ref : int
            Reference ell for normalisation.
        nu_ref_dust : float
            Dust reference frequency.
        nu_ref_sync : float
            Synchrotron reference frequency.
        T_GALDUST : float
            Temperature of the CIB.
        dust_amp_param : str
            The name of the dust amplitude parameter.
        dust_alpha_param : str
            The name of the dust power law index parameter.
        dust_beta_param : str
            The name of the dust frequency scaling parameter.
        sync_amp_param : str
            The name of the sync amplitude parameter.
        sync_alpha_param : str
            The name of the sync power law index parameter.
        sync_beta_param : str
            The name of the sync frequency scaling parameter.
        correlation_param : str
            Name of the correlation parameter.

        Returns
        -------
        GalacticDustSynchrotronCorrelationBandPass
            A new instance of the class.
        """

        super().__init__(
            ells=ells,
            spec_order=spec_order,
            bandpass_info=bandpass_info,
            ell_ref=ell_ref,
            nu_ref={"dust": nu_ref_dust, "sync": nu_ref_sync},
            descriptor=descriptor,
            param_names=[
                correlation_param,
                dust_amp_param,
                dust_alpha_param,
                dust_beta_param,
                sync_amp_param,
                sync_alpha_param,
                sync_beta_param,
            ],
        )

        self.T_dust = T_GALDUST
        self.dust_amp_param = dust_amp_param
        self.dust_alpha_param = dust_alpha_param
        self.dust_beta_param = dust_beta_param

        self.sync_amp_param = sync_amp_param
        self.sync_alpha_param = sync_alpha_param
        self.sync_beta_param = sync_beta_param

        self.correlation_param = correlation_param

        self.affected_specs = affected_specs
        self.spec_mask = jnp.asarray(
            [spec in self.affected_specs for spec in self.spec_order]
        )

        # Turn spectrum mask into a full mask
        self.full_mask = jnp.asarray(
            jnp.repeat(self.spec_mask, len(self.ells)), dtype=float
        )

    @partial(jit, static_argnums=(0,))
    def output(self, sample_params):
        """
        Return foreground spectrum.

        Arguments
        -------
        sampled_params : dict
            Dictionary of nuisance parameter values.

        Returns
        -------
        array, float
            Foreground spectrum.
        """

        # amplitude part (frequency scaling)
        amp_vals_1 = jnp.array(
            [
                candl.transformations.common.dust_frequency_scaling_bandpass(
                    sample_params[self.dust_beta_param],
                    self.T_dust,
                    self.nu_ref["dust"],
                    self.bandpass_info[i][0].nu_spacing,
                    self.bandpass_info[i][0].nu_vals,
                    self.bandpass_info[i][0].bandpass_vals,
                    self.bandpass_info[i][0].thermo_conv[self.nu_ref["dust"]],
                )
                * candl.transformations.common.sync_frequency_scaling_bandpass(
                    sample_params[self.sync_beta_param],
                    self.nu_ref["sync"],
                    self.bandpass_info[i][1].nu_spacing,
                    self.bandpass_info[i][1].nu_vals,
                    self.bandpass_info[i][1].bandpass_vals,
                    self.bandpass_info[i][1].thermo_conv[self.nu_ref["sync"]],
                )
                for i in range(self.N_spec)
            ]
        )

        amp_vals_2 = jnp.array(
            [
                candl.transformations.common.dust_frequency_scaling_bandpass(
                    sample_params[self.dust_beta_param],
                    self.T_dust,
                    self.nu_ref["dust"],
                    self.bandpass_info[i][1].nu_spacing,
                    self.bandpass_info[i][1].nu_vals,
                    self.bandpass_info[i][1].bandpass_vals,
                    self.bandpass_info[i][1].thermo_conv[self.nu_ref["dust"]],
                )
                * candl.transformations.common.sync_frequency_scaling_bandpass(
                    sample_params[self.sync_beta_param],
                    self.nu_ref["sync"],
                    self.bandpass_info[i][0].nu_spacing,
                    self.bandpass_info[i][0].nu_vals,
                    self.bandpass_info[i][0].bandpass_vals,
                    self.bandpass_info[i][0].thermo_conv[self.nu_ref["sync"]],
                )
                for i in range(self.N_spec)
            ]
        )

        amp_vals = amp_vals_1 + amp_vals_2

        amp_vals *= sample_params[self.correlation_param] * jnp.sqrt(
            sample_params[self.dust_amp_param] * sample_params[self.sync_amp_param]
        )
        tiled_amp_vals = jnp.repeat(amp_vals, len(self.ells))

        # ell part
        ell_dependence = (self.ells / self.ell_ref) ** (
            (
                sample_params[self.dust_alpha_param]
                + 2.0
                + sample_params[self.sync_alpha_param]
            )
            / 2.0
        )
        tiled_ell_dependence = jnp.tile(
            ell_dependence, self.N_spec
        )  # tiled ell dependence

        # Complete foreground contribution and mask down
        fg_pow = self.full_mask * tiled_amp_vals * tiled_ell_dependence
        return fg_pow

    @partial(jit, static_argnums=(0,))
    def transform(self, Dls, sample_params):
        """
        Transform spectrum by adding foreground component (result of output method).

        Arguments
        -------
        Dls : array
            Dls to transform.
        sampled_params : dict
            Dictionary of nuisance parameter values.

        Returns
        -------
        array, float
            Transformed spectrum.
        """

        return Dls + self.output(sample_params)


class GalacticDustFixedAlphaBandPass(
    candl.transformations.abstract_base.ForegroundBandPass
):
    """
    Dusty foreground with modified black-body frequency scaling with integral over band pass with a power law ell
    power spectrum.
    $ A * f(beta, bdp_1) * f(beta, bdp_2) * \left( \ell / \ell_{ref} \right)^(-0.4) $
    where:
        A : amplitude
        ell_{ref} : reference ell
        beta : frequency scaling parameter
        f(beta, bdp_1) : is the frequency scaling given the band pass

    User required arguments in data set yaml file
    ---------
    ell_ref : float
        Reference ell.
    nu_ref : float
        Reference frequency.
    T_GALDUST : float
        Temperature of the dust.
    amp_param : str
        The name of the amplitude parameter.
    beta_param : str
        The name of the frequency scaling parameter.
    affected_specs : str
        List of spectrum identifiers the transformation is applied to.

    Example yaml block to add polarised galactic dust to all BB spectra:
    - Module: "candl.transformations.common.GalacticDustBandPass"
      amp_param: "BB_GalDust_BDP_Amp"
      beta_param: "BB_GalDust_BDP_Beta"
      nu_ref: 353
      affected_specs: ["BB 90x90", "BB 90x150", "BB 90x220", "BB 150x150", "BB 150x220", "BB 220x220"]
      ell_ref: 80
      T_GALDUST: 19.6
      descriptor: "BB Polarised Galactic Dust (Bandpass)"

    Methods
    ---------
    __init__ :
        initialises an instance of the class.
    output :
        gives the additive foreground contribution.
    transform :
        transforms an input spectrum.

    Attributes
    -------
    ells : array (float)
        The ell range the transformation acts on.
    descriptor : str
        A short descriptor.
    par_names : list
        Names of parameters involved in transformation.
    spec_order : array (str)
        Identifiers of spectra in the order in which spectra are handled in the long data vector.
    bandpass_info : list
        List of lists, where each sublist contains the two candl.transformations.abstract_base.BandPass instances for the two
        frequencies involved.
    ell_ref : int
        Reference ell for normalisation.
    nu_ref : float
        Reference frequency.
    N_spec : int
        The total number of spectra in the long data vector.
    affected_specs : list (str)
        List of the spectra to apply this foreground to.
    spec_mask : array (int)
        Masks which spectra of the long data vector are affected by the transformation.
    full_mask : array (int)
        Masks which elements of the long data vector are affected by the transformation.
    T_dust : float
        Dust temperature.
    amp_param : str
        The name of the amplitude parameter.
    beta_param : str
        The name of the frequency scaling parameter.

    See also
    -------
    dust_frequency_scaling_bandpass
    """

    def __init__(
        self,
        ells,
        spec_order,
        bandpass_info,
        affected_specs,
        amp_param,
        beta_param,
        ell_ref,
        nu_ref,
        T_GALDUST,
        descriptor="Galactic Dust (Band pass)",
    ):
        """
        Initialise a new instance of the GalacticDustBandPass class.

        Arguments
        -------
        ells : array (float)
            The ell range the transformation acts on.
        descriptor : str
            A short descriptor.
        bandpass_info : list
            List of lists, where each sublist contains the two candl.transformations.abstract_base.BandPass instances for the two
            frequencies involved.
        spec_order : array (str)
            Identifiers of spectra in the order in which spectra are handled in the long data vector.
        affected_specs : list (str)
            List of the spectra to apply this foreground to.
        ell_ref : int
            Reference ell for normalisation.
        nu_ref : float
            Reference frequency.
        T_GALDUST : float
            Temperature of the CIB.
        amp_param : str
            The name of the amplitude parameter.
        beta_param : str
            The name of the frequency scaling parameter.

        Returns
        -------
        GalacticDustFixedAlphaBandPass
            A new instance of the class.
        """

        super().__init__(
            ells=ells,
            spec_order=spec_order,
            bandpass_info=bandpass_info,
            ell_ref=ell_ref,
            nu_ref=nu_ref,
            descriptor=descriptor,
            param_names=[amp_param, beta_param],
        )

        self.T_dust = T_GALDUST
        self.amp_param = amp_param
        self.beta_param = beta_param

        self.affected_specs = affected_specs
        self.spec_mask = jnp.asarray(
            [spec in self.affected_specs for spec in self.spec_order]
        )

        # Turn spectrum mask into a full mask
        self.full_mask = jnp.asarray(
            jnp.repeat(self.spec_mask, len(self.ells)), dtype=float
        )

    @partial(jit, static_argnums=(0,))
    def output(self, sample_params):
        """
        Return foreground spectrum.

        Arguments
        -------
        sampled_params : dict
            Dictionary of nuisance parameter values.

        Returns
        -------
        array, float
            Foreground spectrum.
        """

        # amplitude part (frequency scaling)
        amp_vals = jnp.array(
            [
                candl.transformations.common.dust_frequency_scaling_bandpass(
                    sample_params[self.beta_param],
                    self.T_dust,
                    self.nu_ref,
                    self.bandpass_info[i][0].nu_spacing,
                    self.bandpass_info[i][0].nu_vals,
                    self.bandpass_info[i][0].bandpass_vals,
                    self.bandpass_info[i][0].thermo_conv[self.nu_ref],
                )
                * candl.transformations.common.dust_frequency_scaling_bandpass(
                    sample_params[self.beta_param],
                    self.T_dust,
                    self.nu_ref,
                    self.bandpass_info[i][1].nu_spacing,
                    self.bandpass_info[i][1].nu_vals,
                    self.bandpass_info[i][1].bandpass_vals,
                    self.bandpass_info[i][1].thermo_conv[self.nu_ref],
                )
                for i in range(self.N_spec)
            ]
        )

        amp_vals *= sample_params[self.amp_param]
        tiled_amp_vals = jnp.repeat(amp_vals, len(self.ells))

        # ell part
        ell_dependence = (self.ells / self.ell_ref) ** (-0.4)
        tiled_ell_dependence = jnp.tile(
            ell_dependence, self.N_spec
        )  # tiled ell dependence

        # Complete foreground contribution and mask down
        fg_pow = self.full_mask * tiled_amp_vals * tiled_ell_dependence
        return fg_pow

    @partial(jit, static_argnums=(0,))
    def transform(self, Dls, sample_params):
        """
        Transform spectrum by adding foreground component (result of output method).

        Arguments
        -------
        Dls : array
            Dls to transform.
        sampled_params : dict
            Dictionary of nuisance parameter values.

        Returns
        -------
        array, float
            Transformed spectrum.
        """

        return Dls + self.output(sample_params)


class SynchrotronFixedAlphaBandPass(
    candl.transformations.abstract_base.ForegroundBandPass
):
    """
    Synchrotron foreground with modified power law frequency scaling with integral over band pass with a power law ell
    power spectrum.
    $ A * f(beta, bdp_1) * f(beta, bdp_2) * \left( \ell / \ell_{ref} \right)^(-0.6) $
    where:
        A : amplitude
        ell_{ref} : reference ell
        beta : frequency scaling parameter
        f(beta, bdp_1) : is the frequency scaling given the band pass

    User required arguments in data set yaml file
    ---------
    ell_ref : float
        Reference ell.
    nu_ref : float
        Reference frequency.
    amp_param : str
        The name of the amplitude parameter.
    beta_param : str
        The name of the frequency scaling parameter.
    affected_specs : str
        List of spectrum identifiers the transformation is applied to.

    Example yaml block to add synchrotron power to all BB spectra:
    - Module: "candl.transformations.common.SynchrotronBandPass"
      amp_param: "BB_Sync_BDP_Amp"
      beta_param: "BB_Sync_BDP_Beta"
      nu_ref: 150
      affected_specs: ["BB 90x90", "BB 90x150", "BB 90x220", "BB 150x150", "BB 150x220", "BB 220x220"]
      ell_ref: 80
      descriptor: "BB Synchrotron (Bandpass)"

    Methods
    ---------
    __init__ :
        initialises an instance of the class.
    output :
        gives the additive foreground contribution.
    transform :
        transforms an input spectrum.

    Attributes
    -------
    ells : array (float)
        The ell range the transformation acts on.
    descriptor : str
        A short descriptor.
    par_names : list
        Names of parameters involved in transformation.
    spec_order : array (str)
        Identifiers of spectra in the order in which spectra are handled in the long data vector.
    bandpass_info : list
        List of lists, where each sublist contains the two candl.transformations.abstract_base.BandPass instances for the two
        frequencies involved.
    ell_ref : int
        Reference ell for normalisation.
    nu_ref : float
        Reference frequency.
    N_spec : int
        The total number of spectra in the long data vector.
    affected_specs : list (str)
        List of the spectra to apply this foreground to.
    spec_mask : array (int)
        Masks which spectra of the long data vector are affected by the transformation.
    full_mask : array (int)
        Masks which elements of the long data vector are affected by the transformation.
    amp_param : str
        The name of the amplitude parameter.
    beta_param : str
        The name of the frequency scaling parameter.

    See also
    -------
    sync_frequency_scaling_bandpass
    """

    def __init__(
        self,
        ells,
        spec_order,
        bandpass_info,
        affected_specs,
        amp_param,
        beta_param,
        ell_ref,
        nu_ref,
        descriptor="Synchrotron (Band pass)",
    ):
        """
        Initialise a new instance of the SynchrotronBandPass class.

        Arguments
        -------
        ells : array (float)
            The ell range the transformation acts on.
        descriptor : str
            A short descriptor.
        bandpass_info : list
            List of lists, where each sublist contains the two candl.transformations.abstract_base.BandPass instances for the two
            frequencies involved.
        spec_order : array (str)
            Identifiers of spectra in the order in which spectra are handled in the long data vector.
        affected_specs : list (str)
            List of the spectra to apply this foreground to.
        ell_ref : int
            Reference ell for normalisation.
        nu_ref : float
            Reference frequency.
        amp_param : str
            The name of the amplitude parameter.
        beta_param : str
            The name of the frequency scaling parameter.

        Returns
        -------
        SynchrotronFixedAlphaBandPass
            A new instance of the class.
        """

        super().__init__(
            ells=ells,
            spec_order=spec_order,
            bandpass_info=bandpass_info,
            ell_ref=ell_ref,
            nu_ref=nu_ref,
            descriptor=descriptor,
            param_names=[amp_param, beta_param],
        )

        self.amp_param = amp_param
        self.beta_param = beta_param

        self.affected_specs = affected_specs
        self.spec_mask = jnp.asarray(
            [spec in self.affected_specs for spec in self.spec_order]
        )

        # Turn spectrum mask into a full mask
        self.full_mask = jnp.asarray(
            jnp.repeat(self.spec_mask, len(self.ells)), dtype=float
        )

    @partial(jit, static_argnums=(0,))
    def output(self, sample_params):
        """
        Return foreground spectrum.

        Arguments
        -------
        sampled_params : dict
            Dictionary of nuisance parameter values.

        Returns
        -------
        array, float
            Foreground spectrum.
        """

        # amplitude part (frequency scaling)
        amp_vals = jnp.array(
            [
                candl.transformations.common.sync_frequency_scaling_bandpass(
                    sample_params[self.beta_param],
                    self.nu_ref,
                    self.bandpass_info[i][0].nu_spacing,
                    self.bandpass_info[i][0].nu_vals,
                    self.bandpass_info[i][0].bandpass_vals,
                    self.bandpass_info[i][0].thermo_conv[self.nu_ref],
                )
                * candl.transformations.common.sync_frequency_scaling_bandpass(
                    sample_params[self.beta_param],
                    self.nu_ref,
                    self.bandpass_info[i][1].nu_spacing,
                    self.bandpass_info[i][1].nu_vals,
                    self.bandpass_info[i][1].bandpass_vals,
                    self.bandpass_info[i][1].thermo_conv[self.nu_ref],
                )
                for i in range(self.N_spec)
            ]
        )

        amp_vals *= sample_params[self.amp_param]
        tiled_amp_vals = jnp.repeat(amp_vals, len(self.ells))

        # ell part
        ell_dependence = (self.ells / self.ell_ref) ** (-0.6)
        tiled_ell_dependence = jnp.tile(
            ell_dependence, self.N_spec
        )  # tiled ell dependence

        # Complete foreground contribution and mask down
        fg_pow = self.full_mask * tiled_amp_vals * tiled_ell_dependence
        return fg_pow

    @partial(jit, static_argnums=(0,))
    def transform(self, Dls, sample_params):
        """
        Transform spectrum by adding foreground component (result of output method).

        Arguments
        -------
        Dls : array
            Dls to transform.
        sampled_params : dict
            Dictionary of nuisance parameter values.

        Returns
        -------
        array, float
            Transformed spectrum.
        """

        return Dls + self.output(sample_params)


class GalacticDustSynchrotronCorrelationFixedAlphaBandPass(
    candl.transformations.abstract_base.ForegroundBandPass
):
    """
    Correlation between galactic dust and synchrotron.
    Uses geometric mean of GalacticDustFixedAlphaBandPass and SynchrotronFixedAlphaBandPass, intended to be used in conjunction with these
    two classes.
    Note: assumes that both have the same ell_ref!

    User required arguments in data set yaml file
    ---------
    ell_ref : float
        Reference ell.
    nu_ref_dust : float
        Dust reference frequency.
    nu_ref_sync : float
        Synchrotron reference frequency.
    correlation_param : str
        Name of the correlation parameter.
    T_GALDUST : float
        Temperature of the dust.
    dust_amp_param : str
        The name of the dust amplitude parameter.
    dust_beta_param : str
        The name of the dust frequency scaling parameter.
    affected_specs : str
        List of spectrum identifiers the transformation is applied to.
    sync_amp_param : str
        The name of the synchrotron amplitude parameter.
    sync_beta_param : str
        The name of the synchrotron frequency scaling parameter.


    Example yaml block to add correlation to BB spectra:
    - Module: "candl.transformations.low_ell_BB.GalacticDustSynchrotronCorrelationFixedAlphaBandPass"
      correlation_param: "BB_rho_dust_sync"
      dust_amp_param: "BB_GalDust_BDP_Amp"
      dust_beta_param: "BB_GalDust_BDP_Beta"
      sync_amp_param: "BB_Sync_BDP_Amp"
      sync_beta_param: "BB_Sync_BDP_Beta"
      nu_ref_dust: 353
      nu_ref_sync: 23
      affected_specs: ["BB 90x90", "BB 90x150", "BB 150x150"]
      ell_ref: 80
      T_GALDUST: 19.6
      descriptor: "BB Dust Synchrotron Correlation (Bandpass)"

    Methods
    ---------
    __init__ :
        initialises an instance of the class.
    output :
        gives the additive foreground contribution.
    transform :
        transforms an input spectrum.

    Attributes
    -------
    ells : array (float)
        The ell range the transformation acts on.
    descriptor : str
        A short descriptor.
    par_names : list
        Names of parameters involved in transformation.
    spec_order : array (str)
        Identifiers of spectra in the order in which spectra are handled in the long data vector.
    bandpass_info : list
        List of lists, where each sublist contains the two candl.transformations.abstract_base.BandPass instances for the two
        frequencies involved.
    ell_ref : int
        Reference ell for normalisation.
    nu_ref_dust : float
        Dust reference frequency.
    nu_ref_sync : float
        Synchrotron reference frequency.
    N_spec : int
        The total number of spectra in the long data vector.
    affected_specs : list (str)
        List of the spectra to apply this foreground to.
    spec_mask : array (int)
        Masks which spectra of the long data vector are affected by the transformation.
    full_mask : array (int)
        Masks which elements of the long data vector are affected by the transformation.
    T_dust : float
        Dust temperature.
    dust_amp_param : str
        The name of the dust amplitude parameter.
    dust_beta_param : str
        The name of the dust frequency scaling parameter.
    sync_amp_param : str
        The name of the sync amplitude parameter.
    sync_beta_param : str
        The name of the sync frequency scaling parameter.
    correlation_param : str
        Name of the correlation parameter.

    See also
    -------
    dust_frequency_scaling_bandpass
    """

    def __init__(
        self,
        ells,
        spec_order,
        bandpass_info,
        affected_specs,
        correlation_param,
        dust_amp_param,
        dust_beta_param,
        sync_amp_param,
        sync_beta_param,
        ell_ref,
        nu_ref_dust,
        nu_ref_sync,
        T_GALDUST,
        descriptor="Dust Synchrotron Correlation (Band pass)",
    ):
        """
        Initialise a new instance of the GalacticDustSynchrotronCorrelationFixedAlphaBandPass class.

        Arguments
        -------
        ells : array (float)
            The ell range the transformation acts on.
        descriptor : str
            A short descriptor.
        bandpass_info : list
            List of lists, where each sublist contains the two candl.transformations.abstract_base.BandPass instances for the two
            frequencies involved.
        spec_order : array (str)
            Identifiers of spectra in the order in which spectra are handled in the long data vector.
        affected_specs : list (str)
            List of the spectra to apply this foreground to.
        ell_ref : int
            Reference ell for normalisation.
        nu_ref_dust : float
            Dust reference frequency.
        nu_ref_sync : float
            Synchrotron reference frequency.
        T_GALDUST : float
            Temperature of the CIB.
        dust_amp_param : str
            The name of the dust amplitude parameter.
        dust_beta_param : str
            The name of the dust frequency scaling parameter.
        sync_amp_param : str
            The name of the sync amplitude parameter.
        sync_beta_param : str
            The name of the sync frequency scaling parameter.
        correlation_param : str
            Name of the correlation parameter.

        Returns
        -------
        GalacticDustSynchrotronCorrelationFixedAlphaBandPass
            A new instance of the class.
        """

        super().__init__(
            ells=ells,
            spec_order=spec_order,
            bandpass_info=bandpass_info,
            ell_ref=ell_ref,
            nu_ref={"dust": nu_ref_dust, "sync": nu_ref_sync},
            descriptor=descriptor,
            param_names=[
                correlation_param,
                dust_amp_param,
                dust_beta_param,
                sync_amp_param,
                sync_beta_param,
            ],
        )

        self.T_dust = T_GALDUST
        self.dust_amp_param = dust_amp_param
        self.dust_beta_param = dust_beta_param

        self.sync_amp_param = sync_amp_param
        self.sync_beta_param = sync_beta_param

        self.correlation_param = correlation_param

        self.affected_specs = affected_specs
        self.spec_mask = jnp.asarray(
            [spec in self.affected_specs for spec in self.spec_order]
        )

        # Turn spectrum mask into a full mask
        self.full_mask = jnp.asarray(
            jnp.repeat(self.spec_mask, len(self.ells)), dtype=float
        )

    @partial(jit, static_argnums=(0,))
    def output(self, sample_params):
        """
        Return foreground spectrum.

        Arguments
        -------
        sampled_params : dict
            Dictionary of nuisance parameter values.

        Returns
        -------
        array, float
            Foreground spectrum.
        """

        # amplitude part (frequency scaling)
        amp_vals_1 = jnp.array(
            [
                candl.transformations.common.dust_frequency_scaling_bandpass(
                    sample_params[self.dust_beta_param],
                    self.T_dust,
                    self.nu_ref["dust"],
                    self.bandpass_info[i][0].nu_spacing,
                    self.bandpass_info[i][0].nu_vals,
                    self.bandpass_info[i][0].bandpass_vals,
                    self.bandpass_info[i][0].thermo_conv[self.nu_ref["dust"]],
                )
                * candl.transformations.common.sync_frequency_scaling_bandpass(
                    sample_params[self.sync_beta_param],
                    self.nu_ref["sync"],
                    self.bandpass_info[i][1].nu_spacing,
                    self.bandpass_info[i][1].nu_vals,
                    self.bandpass_info[i][1].bandpass_vals,
                    self.bandpass_info[i][1].thermo_conv[self.nu_ref["sync"]],
                )
                for i in range(self.N_spec)
            ]
        )

        amp_vals_2 = jnp.array(
            [
                candl.transformations.common.dust_frequency_scaling_bandpass(
                    sample_params[self.dust_beta_param],
                    self.T_dust,
                    self.nu_ref["dust"],
                    self.bandpass_info[i][1].nu_spacing,
                    self.bandpass_info[i][1].nu_vals,
                    self.bandpass_info[i][1].bandpass_vals,
                    self.bandpass_info[i][1].thermo_conv[self.nu_ref["dust"]],
                )
                * candl.transformations.common.sync_frequency_scaling_bandpass(
                    sample_params[self.sync_beta_param],
                    self.nu_ref["sync"],
                    self.bandpass_info[i][0].nu_spacing,
                    self.bandpass_info[i][0].nu_vals,
                    self.bandpass_info[i][0].bandpass_vals,
                    self.bandpass_info[i][0].thermo_conv[self.nu_ref["sync"]],
                )
                for i in range(self.N_spec)
            ]
        )

        amp_vals = amp_vals_1 + amp_vals_2

        amp_vals *= sample_params[self.correlation_param] * jnp.sqrt(
            sample_params[self.dust_amp_param] * sample_params[self.sync_amp_param]
        )
        tiled_amp_vals = jnp.repeat(amp_vals, len(self.ells))

        # ell part
        ell_dependence = (self.ells / self.ell_ref) ** (-0.5)
        tiled_ell_dependence = jnp.tile(
            ell_dependence, self.N_spec
        )  # tiled ell dependence

        # Complete foreground contribution and mask down
        fg_pow = self.full_mask * tiled_amp_vals * tiled_ell_dependence
        return fg_pow

    @partial(jit, static_argnums=(0,))
    def transform(self, Dls, sample_params):
        """
        Transform spectrum by adding foreground component (result of output method).

        Arguments
        -------
        Dls : array
            Dls to transform.
        sampled_params : dict
            Dictionary of nuisance parameter values.

        Returns
        -------
        array, float
            Transformed spectrum.
        """

        return Dls + self.output(sample_params)


# --------------------------------------#
# DUST AND SYNCHROTRON WITH BAND PASS INTEGRALS - DECORRELATION ACROSS FREQUENCIES
# --------------------------------------#


class GalacticDustDecorrelationBandPass(GalacticDustBandPass):
    """
    Polarised galactic dust including decorrelation for cross-frequency spectra.
    Slightly modified version of GalacticDustBandPass (see for more extensive documentation).
    Requires additionally "decorr_param", "nu_pivot" set.
    Argument "ell_scaling" is optional: default is flat ell-scaling, other options are "linear" and "quadratic".

    Example yaml block to add polarised galactic dust to all BB spectra:
    - Module: "candl.transformations.low_ell_BB.GalacticDustDecorrelationBandPass"
      decorr_param: "BB_GalDust_BDP_Decorr"
      amp_param: "BB_GalDust_BDP_Amp"
      alpha_param: "BB_GalDust_BDP_Alpha"
      beta_param: "BB_GalDust_BDP_Beta"
      nu_ref: 353
      nu_pivot: [353, 217]
      affected_specs: ["BB 90x90", "BB 90x150", "BB 150x150"]
      ell_ref: 80
      T_GALDUST: 19.6
      descriptor: "BB Polarised Galactic Dust (Decorrelation, Bandpass)"

    Follows implementation in BK Likelihood, see BK15 (https://arxiv.org/pdf/1810.05216.pdf) Appendix F for details.
    Not extended for unphysical values of decorr_param.
    """

    def __init__(
        self,
        ells,
        spec_order,
        bandpass_info,
        affected_specs,
        decorr_param,
        amp_param,
        alpha_param,
        beta_param,
        ell_ref,
        nu_ref,
        nu_pivot,
        T_GALDUST,
        ell_scaling=None,
        descriptor="Galactic Dust (Decorrelation, Band pass)",
    ):

        super().__init__(
            ells=ells,
            spec_order=spec_order,
            bandpass_info=bandpass_info,
            affected_specs=affected_specs,
            amp_param=amp_param,
            alpha_param=alpha_param,
            beta_param=beta_param,
            ell_ref=ell_ref,
            nu_ref=nu_ref,
            T_GALDUST=T_GALDUST,
            descriptor=descriptor,
        )

        self.decorr_param = decorr_param

        self.nu_pivot = nu_pivot

        self.ell_scaling = ell_scaling
        self.ell_part = 1.0
        if self.ell_scaling == "linear":
            self.ell_part = jnp.tile(self.ells, self.N_spec)
        elif self.ell_scaling == "quadratic":
            self.ell_part = jnp.tile(self.ells**2, self.N_spec)

        # Make mask of auto-frequency spectra (decorrelation does not apply to them)
        self.auto_spec_mask = jnp.asarray(
            [len(set(s[3:].split("x"))) == 1 for s in self.spec_order]
        )
        self.full_auto_mask = jnp.repeat(self.auto_spec_mask, len(self.ells))

    def output(self, sample_params):

        # Get regular dust power
        fg_pow = super().output(sample_params)

        # Calculate decorrelation
        amp_vals = jnp.array(
            [
                jnp.log(
                    self.bandpass_info[i][0].central_nu
                    / self.bandpass_info[i][1].central_nu
                )
                ** 2
                / jnp.log(self.nu_pivot[0] / self.nu_pivot[1]) ** 2
                for i in range(self.N_spec)
            ]
        )
        amp_vals_tiled = jnp.repeat(amp_vals, len(self.ells))
        delta_prime = jnp.exp(
            jnp.log(sample_params[self.decorr_param]) * amp_vals_tiled * self.ell_part
        )
        delta_prime = jax_optional_set_element(delta_prime, self.full_auto_mask, 1.0)

        return delta_prime * fg_pow


class GalacticDustFixedAlphaDecorrelationBandPass(GalacticDustFixedAlphaBandPass):
    """
    Slightly modified version of GalacticDustDecorrelationBandPass to fix the power law index (as in GalacticDustFixedAlphaBandPass).
    """

    def __init__(
        self,
        ells,
        spec_order,
        bandpass_info,
        affected_specs,
        decorr_param,
        amp_param,
        beta_param,
        ell_ref,
        nu_ref,
        nu_pivot,
        T_GALDUST,
        ell_scaling=None,
        descriptor="Galactic Dust (Decorrelation, Band pass)",
    ):

        super().__init__(
            ells=ells,
            spec_order=spec_order,
            bandpass_info=bandpass_info,
            affected_specs=affected_specs,
            amp_param=amp_param,
            beta_param=beta_param,
            ell_ref=ell_ref,
            nu_ref=nu_ref,
            T_GALDUST=T_GALDUST,
            descriptor=descriptor,
        )

        self.decorr_param = decorr_param

        self.nu_pivot = nu_pivot

        self.ell_scaling = ell_scaling
        self.ell_part = 1.0
        if self.ell_scaling == "linear":
            self.ell_part = jnp.tile(self.ells, self.N_spec)
        elif self.ell_scaling == "quadratic":
            self.ell_part = jnp.tile(self.ells**2, self.N_spec)

        # Make mask of auto-frequency spectra (decorrelation does not apply to them)
        self.auto_spec_mask = jnp.asarray(
            [len(set(s[3:].split("x"))) == 1 for s in self.spec_order]
        )
        self.full_auto_mask = jnp.repeat(self.auto_spec_mask, len(self.ells))

    def output(self, sample_params):

        # Get regular dust power
        fg_pow = super().output(sample_params)

        # Calculate decorrelation
        amp_vals = jnp.array(
            [
                jnp.log(
                    self.bandpass_info[i][0].central_nu
                    / self.bandpass_info[i][1].central_nu
                )
                ** 2
                / jnp.log(self.nu_pivot[0] / self.nu_pivot[1]) ** 2
                for i in range(self.N_spec)
            ]
        )
        amp_vals_tiled = jnp.repeat(amp_vals, len(self.ells))
        delta_prime = jnp.exp(
            jnp.log(sample_params[self.decorr_param]) * amp_vals_tiled * self.ell_part
        )
        delta_prime = jax_optional_set_element(delta_prime, self.full_auto_mask, 1.0)

        return delta_prime * fg_pow



"""
A series of useful constants.

Overview
-----------

* :const:`T_CMB` - CMB temperature
* :const:`PLANCK_h` - Planck's constant
* :const:`BOLTZMANN_kB` - Boltzmann constant
* :const:`GHz_KELVIN` - Useful shortcut for black-body related functions
"""

# Physical constants
T_CMB = 2.72548  # CMB temperature
PLANCK_h = 6.62606957e-34  # Planck's constant
BOLTZMANN_kB = 1.3806488e-23  # Boltzmann constant
GHz_KELVIN = (
    PLANCK_h / BOLTZMANN_kB * 1e9
)  # Useful shortcut for black-body related functions
